import os

import numpy as np
import pytest

from docarray import Document
from docarray.document.generators import from_files
from docarray.helper import __windows__

cur_dir = os.path.dirname(os.path.abspath(__file__))


def test_video_convert_pipe(pytestconfig, tmpdir):
    num_d = 0
    fname = str(tmpdir / f'tmp{num_d}.mp4')
    d = Document(uri=os.path.join(cur_dir, 'toydata/mov_bbb.mp4'))
    d.load_uri_to_video_blob()
    d.save_video_blob_to_file(fname)
    assert os.path.exists(fname)


def test_audio_convert_pipe(pytestconfig, tmpdir):
    num_d = 0
    for d in from_files(f'{cur_dir}/toydata/*.wav'):
        fname = str(tmpdir / f'tmp{num_d}.wav')
        d.load_uri_to_audio_blob()
        d.blob = d.blob[::-1]
        d.save_audio_blob_to_file(fname)
        assert os.path.exists(fname)
        num_d += 1
    assert num_d


def test_image_convert_pipe(pytestconfig):
    for d in from_files(f'{pytestconfig.rootdir}/.github/**/*.png'):
        (
            d.load_uri_to_image_blob()
            .convert_uri_to_datauri()
            .set_image_blob_shape((64, 64))
            .set_image_blob_normalization()
            .set_image_blob_channel_axis(-1, 0)
        )
        assert d.blob.shape == (3, 64, 64)
        assert d.uri


def test_uri_to_blob():
    doc = Document(uri=os.path.join(cur_dir, 'toydata/test.png'))
    doc.load_uri_to_image_blob()
    assert isinstance(doc.blob, np.ndarray)
    assert doc.blob.shape == (85, 152, 3)  # h,w,c
    assert doc.mime_type == 'image/png'


def test_datauri_to_blob():
    doc = Document(uri=os.path.join(cur_dir, 'toydata/test.png'))
    doc.convert_uri_to_datauri()
    assert not doc.blob
    assert doc.mime_type == 'image/png'


def test_buffer_to_blob():
    doc = Document(uri=os.path.join(cur_dir, 'toydata/test.png'))
    doc.load_uri_to_buffer()
    doc.convert_buffer_to_image_blob()
    assert isinstance(doc.blob, np.ndarray)
    assert doc.mime_type == 'image/png'
    assert doc.blob.shape == (85, 152, 3)  # h,w,c


def test_convert_buffer_to_blob():
    rand_state = np.random.RandomState(0)
    array = rand_state.random([10, 10])
    doc = Document(content=array.tobytes())
    assert doc.content_type == 'buffer'
    intialiazed_buffer = doc.buffer

    doc.convert_buffer_to_blob()
    assert doc.content_type == 'blob'
    converted_buffer_in_one_of = doc.buffer
    assert intialiazed_buffer != converted_buffer_in_one_of
    np.testing.assert_almost_equal(doc.content.reshape([10, 10]), array)


@pytest.mark.parametrize('shape, channel_axis', [((3, 32, 32), 0), ((32, 32, 3), -1)])
def test_image_normalize(shape, channel_axis):
    doc = Document(content=np.random.randint(0, 255, shape, dtype=np.uint8))
    doc.set_image_blob_normalization(channel_axis=channel_axis)
    assert doc.blob.ndim == 3
    assert doc.blob.shape == shape
    assert doc.blob.dtype == np.float32


@pytest.mark.parametrize(
    'arr_size, channel_axis, height, width',
    [
        ([32, 28, 3], -1, 32, 28),  # h, w, c (rgb)
        ([3, 32, 28], 0, 32, 28),  # c, h, w  (rgb)
        ([1, 32, 28], 0, 32, 28),  # c, h, w, (greyscale)
        ([32, 28, 1], -1, 32, 28),  # h, w, c, (greyscale)
    ],
)
def test_convert_image_blob_to_uri(arr_size, channel_axis, width, height):
    doc = Document(content=np.random.randint(0, 255, arr_size))
    assert doc.blob.any()
    assert not doc.uri
    doc.set_image_blob_shape(channel_axis=channel_axis, shape=(width, height))

    doc.convert_image_blob_to_uri(channel_axis=channel_axis)
    assert doc.uri.startswith('data:image/png;base64,')
    assert doc.mime_type == 'image/png'
    assert doc.blob.any()  # assure after conversion blob still exist.


@pytest.mark.xfail(
    condition=__windows__, reason='x-python is not detected on windows CI'
)
@pytest.mark.parametrize(
    'uri, mimetype',
    [
        (__file__, 'text/x-python'),
        ('http://google.com/index.html', 'text/html'),
        ('https://google.com/index.html', 'text/html'),
    ],
)
def test_convert_uri_to_buffer(uri, mimetype):
    d = Document(uri=uri)
    assert not d.buffer
    d.load_uri_to_buffer()
    assert d.buffer
    assert d.mime_type == mimetype


@pytest.mark.parametrize(
    'converter', ['convert_buffer_to_datauri', 'convert_content_to_datauri']
)
def test_convert_buffer_to_uri(converter):
    d = Document(content=open(__file__).read().encode(), mime_type='text/x-python')
    assert d.buffer
    getattr(d, converter)()
    assert d.uri.startswith('data:text/x-python;')


@pytest.mark.parametrize(
    'converter', ['convert_text_to_datauri', 'convert_content_to_datauri']
)
def test_convert_text_to_uri(converter):
    d = Document(content=open(__file__).read())
    assert d.text
    getattr(d, converter)()
    assert d.uri.startswith('data:text/plain;')


@pytest.mark.xfail(
    condition=__windows__, reason='x-python is not detected on windows CI'
)
@pytest.mark.parametrize(
    'uri, mimetype',
    [
        pytest.param(
            __file__,
            'text/x-python',
            marks=pytest.mark.xfail(
                condition=__windows__, reason='x-python is not detected on windows CI'
            ),
        ),
        ('http://google.com/index.html', 'text/html'),
        ('https://google.com/index.html', 'text/html'),
    ],
)
def test_convert_uri_to_text(uri, mimetype):
    doc = Document(uri=uri, mime_type=mimetype)
    doc.load_uri_to_text()
    if mimetype == 'text/html':
        assert '<!doctype html>' in doc.text
    elif mimetype == 'text/x-python':
        text_from_file = open(__file__).read()
        assert doc.text == text_from_file


def test_convert_text_to_uri_and_back():
    text_from_file = open(__file__).read()
    doc = Document(content=text_from_file, mime_type='text/x-python')
    assert doc.text
    assert doc.mime_type == 'text/x-python'
    doc.convert_text_to_datauri()
    doc.load_uri_to_text()
    assert doc.mime_type == 'text/plain'
    assert doc.text == text_from_file


def test_convert_text_diff_encoding(tmpfile):
    otext = 'testä'
    text = otext.encode('iso8859')
    with open(tmpfile, 'wb') as fp:
        fp.write(text)
    with pytest.raises(UnicodeDecodeError):
        d = Document(uri=str(tmpfile)).load_uri_to_text()

    d = Document(uri=str(tmpfile)).load_uri_to_text(charset='iso8859')
    assert d.text == otext

    with open(tmpfile, 'w', encoding='iso8859') as fp:
        fp.write(otext)
    with pytest.raises(UnicodeDecodeError):
        d = Document(uri=str(tmpfile)).load_uri_to_text()

    d = Document(uri=str(tmpfile)).load_uri_to_text(charset='iso8859')
    assert d.text == otext


def test_convert_content_to_uri():
    d = Document(content=np.random.random([10, 10]))
    with pytest.raises(NotImplementedError):
        d.convert_content_to_datauri()


@pytest.mark.parametrize(
    'uri, mimetype',
    [
        (__file__, 'text/x-python'),
        ('http://google.com/index.html', 'text/html'),
        ('https://google.com/index.html', 'text/html'),
    ],
)
def test_convert_uri_to_data_uri(uri, mimetype):
    doc = Document(uri=uri, mime_type=mimetype)
    doc.convert_uri_to_datauri()
    assert doc.uri.startswith(f'data:{mimetype}')
    assert doc.mime_type == mimetype


def test_glb_converters():
    doc = Document(uri=os.path.join(cur_dir, 'toydata/test.glb'))
    doc.load_uri_to_point_cloud_blob(2000)
    assert doc.blob.shape == (2000, 3)



import json
import os
import random

import numpy as np
import pytest

from docarray import DocumentArray, Document


def test_sprite_image_generator(pytestconfig, tmpdir):
    da = DocumentArray.from_files(
        [
            f'{pytestconfig.rootdir}/**/*.png',
            f'{pytestconfig.rootdir}/**/*.jpg',
            f'{pytestconfig.rootdir}/**/*.jpeg',
        ]
    )
    da.plot_image_sprites(tmpdir / 'sprint_da.png')
    assert os.path.exists(tmpdir / 'sprint_da.png')


def da_and_dam():
    embeddings = np.array([[1, 0, 0], [2, 0, 0], [3, 0, 0]])
    doc_array = DocumentArray(
        [
            Document(embedding=x, tags={'label': random.randint(0, 5)})
            for x in embeddings
        ]
    )

    return (doc_array,)


@pytest.mark.parametrize('da', da_and_dam())
def test_plot_embeddings(da):
    p = da.plot_embeddings(start_server=False)
    assert os.path.exists(p)
    assert os.path.exists(os.path.join(p, 'config.json'))
    with open(os.path.join(p, 'config.json')) as fp:
        config = json.load(fp)
        assert len(config['embeddings']) == 1
        assert config['embeddings'][0]['tensorShape'] == list(da.embeddings.shape)


def test_plot_embeddings_same_path(tmpdir):
    da1 = DocumentArray.empty(100)
    da1.embeddings = np.random.random([100, 5])
    p1 = da1.plot_embeddings(start_server=False, path=tmpdir)
    da2 = DocumentArray.empty(768)
    da2.embeddings = np.random.random([768, 5])
    p2 = da2.plot_embeddings(start_server=False, path=tmpdir)
    assert p1 == p2
    assert os.path.exists(p1)
    with open(os.path.join(p1, 'config.json')) as fp:
        config = json.load(fp)
        assert len(config['embeddings']) == 2


def test_summary_homo_hetero():
    da = DocumentArray.empty(100)
    da.get_attributes()
    da.summary()

    da[0].pop('id')
    da.summary()


def test_empty_get_attributes():
    da = DocumentArray.empty(10)
    da[0].pop('id')
    print(da.get_attributes('id'))



from typing import TYPE_CHECKING, Type

if TYPE_CHECKING:
    from ...types import T
    from ...proto.docarray_pb2 import DocumentProto


class ProtobufMixin:
    @classmethod
    def from_protobuf(cls: Type['T'], pb_msg: 'DocumentProto') -> 'T':
        from ...proto.io import parse_proto

        return parse_proto(pb_msg)

    def to_protobuf(self) -> 'DocumentProto':
        from ...proto.io import flush_proto

        return flush_proto(self)



import numpy as np

from docarray import DocumentArray, Document


def random_docs(
    num_docs,
    chunks_per_doc=5,
    embed_dim=10,
    jitter=1,
    start_id=0,
    embedding=True,
    sparse_embedding=False,
    text='hello world',
) -> DocumentArray:
    da = DocumentArray()
    next_chunk_doc_id = start_id + num_docs
    for j in range(num_docs):
        doc_id = str(start_id + j)

        d = Document(id=doc_id)
        d.text = text
        d.tags['id'] = doc_id
        if embedding:
            if sparse_embedding:
                from scipy.sparse import coo_matrix

                d.embedding = coo_matrix(
                    (np.array([1, 1, 1]), (np.array([0, 1, 2]), np.array([1, 2, 1])))
                )
            else:
                d.embedding = np.random.random(
                    [embed_dim + np.random.randint(0, jitter)]
                )

        for _ in range(chunks_per_doc):
            chunk_doc_id = str(next_chunk_doc_id)

            c = Document(id=chunk_doc_id)
            c.text = 'i\'m chunk %s from doc %s' % (chunk_doc_id, doc_id)
            if embedding:
                c.embedding = np.random.random(
                    [embed_dim + np.random.randint(0, jitter)]
                )
            c.tags['parent_id'] = doc_id
            c.tags['id'] = chunk_doc_id
            d.chunks.append(c)
            next_chunk_doc_id += 1

        da.append(d)
    return da



import pytest

from docarray import DocumentArray
from docarray.math.ndarray import to_numpy_array


@pytest.mark.parametrize('n_dim', [2, 4, 100])
@pytest.mark.parametrize('sparse', [True, False])
@pytest.mark.parametrize('metric', ['jaccard', 'cosine'])
def test_feature_hashing(n_dim, sparse, metric):
    da = DocumentArray.empty(3)
    da.texts = ['hello world', 'world, bye', 'hello bye']
    da.apply(lambda d: d.embed_feature_hashing(n_dim=n_dim, sparse=sparse))
    assert da.embeddings.shape == (3, n_dim)
    da.embeddings = to_numpy_array(da.embeddings)
    da.match(da, metric=metric, use_scipy=True)
    result = da['@m', ('id', f'scores__{metric}__value')]
    assert len(result) == 2
    assert result[1][0] == 0.0
    assert result[1][1] > 0.0



import pytest

from docarray import Document, DocumentArray


@pytest.mark.parametrize('da_cls', [DocumentArray])
def test_construct_docarray(da_cls):
    da = da_cls()
    assert len(da) == 0

    da = da_cls(Document())
    assert len(da) == 1

    da = da_cls([Document(), Document()])
    assert len(da) == 2

    da = da_cls((Document(), Document()))
    assert len(da) == 2

    da = da_cls((Document() for _ in range(10)))
    assert len(da) == 10

    da1 = da_cls(da)
    assert len(da1) == 10


@pytest.mark.parametrize('da_cls', [DocumentArray])
@pytest.mark.parametrize('is_copy', [True, False])
def test_docarray_copy_singleton(da_cls, is_copy):
    d = Document()
    da = da_cls(d, copy=is_copy)
    d.id = 'hello'
    if is_copy:
        assert da[0].id != 'hello'
    else:
        assert da[0].id == 'hello'


@pytest.mark.parametrize('da_cls', [DocumentArray])
@pytest.mark.parametrize('is_copy', [True, False])
def test_docarray_copy_da(da_cls, is_copy):
    d1 = Document()
    d2 = Document()
    da = da_cls([d1, d2], copy=is_copy)
    d1.id = 'hello'
    if is_copy:
        assert da[0].id != 'hello'
    else:
        assert da[0].id == 'hello'


@pytest.mark.parametrize('da_cls', [DocumentArray])
@pytest.mark.parametrize('is_copy', [True, False])
def test_docarray_copy_list(da_cls, is_copy):
    d1 = Document()
    d2 = Document()
    da = da_cls([d1, d2], copy=is_copy)
    d1.id = 'hello'
    if is_copy:
        assert da[0].id != 'hello'
    else:
        assert da[0].id == 'hello'



import itertools
import types

import numpy as np
import pytest

from docarray import Document, DocumentArray
from tests import random_docs

# some random prime number for sanity check
num_docs = 7
num_chunks_per_doc = 11
num_matches_per_doc = 3
num_matches_per_chunk = 5


@pytest.fixture
def doc_req():
    """Build a dummy request that has docs"""
    ds = list(random_docs(num_docs, num_chunks_per_doc))
    # add some random matches
    for d in ds:
        for _ in range(num_matches_per_doc):
            d.matches.append(Document(content='hello'))
        for c in d.chunks:
            for _ in range(num_matches_per_chunk):
                c.matches.append(Document(content='world'))
    yield DocumentArray(ds)


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_traverse_type(doc_req, filter_fn):
    ds = doc_req.traverse('r', filter_fn=filter_fn)
    assert isinstance(ds, types.GeneratorType)
    assert isinstance(list(ds)[0], DocumentArray)


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_traverse_root(doc_req, filter_fn):
    ds = list(doc_req.traverse('r', filter_fn=filter_fn))
    assert len(ds) == 1
    assert len(ds[0]) == num_docs


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_traverse_chunk(doc_req, filter_fn):
    ds = list(doc_req.traverse('c', filter_fn=filter_fn))
    assert len(ds) == num_docs
    assert len(ds[0]) == num_chunks_per_doc


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_traverse_root_plus_chunk(doc_req, filter_fn):
    ds = list(doc_req.traverse('c,r', filter_fn=filter_fn))
    assert len(ds) == num_docs + 1
    assert len(ds[0]) == num_chunks_per_doc
    assert len(ds[-1]) == num_docs


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_traverse_chunk_plus_root(doc_req, filter_fn):
    ds = list(doc_req.traverse('r,c', filter_fn=filter_fn))
    assert len(ds) == 1 + num_docs
    assert len(ds[-1]) == num_chunks_per_doc
    assert len(ds[0]) == num_docs


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_traverse_match(doc_req, filter_fn):
    ds = list(doc_req.traverse('m', filter_fn=filter_fn))
    assert len(ds) == num_docs
    assert len(ds[0]) == num_matches_per_doc


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_traverse_match_chunk(doc_req, filter_fn):
    ds = list(doc_req.traverse('cm', filter_fn=filter_fn))
    assert len(ds) == num_docs * num_chunks_per_doc
    assert len(ds[0]) == num_matches_per_chunk


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_traverse_root_match_chunk(doc_req, filter_fn):
    ds = list(doc_req.traverse('r,c,m,cm', filter_fn=filter_fn))
    assert len(ds) == 1 + num_docs + num_docs + num_docs * num_chunks_per_doc


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_traverse_flatten_embedding(doc_req, filter_fn):
    flattened_results = doc_req.traverse_flat('r,c', filter_fn=filter_fn)
    ds = np.stack(flattened_results.get_attributes('embedding'))
    assert ds.shape == (num_docs + num_chunks_per_doc * num_docs, 10)


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_traverse_flatten_root(doc_req, filter_fn):
    ds = list(doc_req.traverse_flat('r', filter_fn=filter_fn))
    assert len(ds) == num_docs


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_traverse_flatten_chunk(doc_req, filter_fn):
    ds = list(doc_req.traverse_flat('c', filter_fn=filter_fn))
    assert len(ds) == num_docs * num_chunks_per_doc


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_traverse_flatten_root_plus_chunk(doc_req, filter_fn):
    ds = list(doc_req.traverse_flat('c,r', filter_fn=filter_fn))
    assert len(ds) == num_docs + num_docs * num_chunks_per_doc


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_traverse_flatten_match(doc_req, filter_fn):
    ds = list(doc_req.traverse_flat('m', filter_fn=filter_fn))
    assert len(ds) == num_docs * num_matches_per_doc


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_traverse_flatten_match_chunk(doc_req, filter_fn):
    ds = list(doc_req.traverse_flat('cm', filter_fn=filter_fn))
    assert len(ds) == num_docs * num_chunks_per_doc * num_matches_per_chunk


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_traverse_flatten_root_match_chunk(doc_req, filter_fn):
    ds = list(doc_req.traverse_flat('r,c,m,cm', filter_fn=filter_fn))
    assert (
        len(ds)
        == num_docs
        + num_chunks_per_doc * num_docs
        + num_matches_per_doc * num_docs
        + num_docs * num_chunks_per_doc * num_matches_per_chunk
    )


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_traverse_flattened_per_path_embedding(doc_req, filter_fn):
    flattened_results = list(doc_req.traverse_flat_per_path('r,c', filter_fn=filter_fn))
    ds = np.stack(flattened_results[0].get_attributes('embedding'))
    assert ds.shape == (num_docs, 10)

    ds = np.stack(flattened_results[1].get_attributes('embedding'))
    assert ds.shape == (num_docs * num_chunks_per_doc, 10)


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_traverse_flattened_per_path_root(doc_req, filter_fn):
    ds = list(doc_req.traverse_flat_per_path('r', filter_fn=filter_fn))
    assert len(ds[0]) == num_docs


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_traverse_flattened_per_path_chunk(doc_req, filter_fn):
    ds = list(doc_req.traverse_flat_per_path('c', filter_fn=filter_fn))
    assert len(ds[0]) == num_docs * num_chunks_per_doc


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_traverse_flattened_per_path_root_plus_chunk(doc_req, filter_fn):
    ds = list(doc_req.traverse_flat_per_path('c,r', filter_fn=filter_fn))
    assert len(ds[0]) == num_docs * num_chunks_per_doc
    assert len(ds[1]) == num_docs


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_traverse_flattened_per_path_match(doc_req, filter_fn):
    ds = list(doc_req.traverse_flat_per_path('m', filter_fn=filter_fn))
    assert len(ds[0]) == num_docs * num_matches_per_doc


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_traverse_flattened_per_path_match_chunk(doc_req, filter_fn):
    ds = list(doc_req.traverse_flat_per_path('cm', filter_fn=filter_fn))
    assert len(ds[0]) == num_docs * num_chunks_per_doc * num_matches_per_chunk


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_traverse_flattened_per_path_root_match_chunk(doc_req, filter_fn):
    ds = list(doc_req.traverse_flat_per_path('r,c,m,cm', filter_fn=filter_fn))
    assert len(ds[0]) == num_docs
    assert len(ds[1]) == num_chunks_per_doc * num_docs
    assert len(ds[2]) == num_matches_per_doc * num_docs
    assert len(ds[3]) == num_docs * num_chunks_per_doc * num_matches_per_chunk


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_docuset_traverse_over_iterator_HACKY(filter_fn):
    # HACKY USAGE DO NOT RECOMMEND: can also traverse over "runtime"-documentarray
    ds = DocumentArray(random_docs(num_docs, num_chunks_per_doc)).traverse(
        'r', filter_fn=filter_fn
    )
    assert len(list(list(ds)[0])) == num_docs

    ds = DocumentArray(random_docs(num_docs, num_chunks_per_doc)).traverse(
        'c', filter_fn=filter_fn
    )
    ds = list(ds)
    assert len(ds) == num_docs
    assert len(ds[0]) == num_chunks_per_doc


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_docuset_traverse_over_iterator_CAVEAT(filter_fn):
    # HACKY USAGE's CAVEAT: but it can not iterate over an iterator twice
    ds = DocumentArray(random_docs(num_docs, num_chunks_per_doc)).traverse(
        'r,c', filter_fn=filter_fn
    )
    # note that random_docs is a generator and can be only used once,
    # therefore whoever comes first wil get iterated, and then it becomes empty
    assert len(list(ds)) == 1 + num_docs

    ds = DocumentArray(random_docs(num_docs, num_chunks_per_doc)).traverse(
        'c,r', filter_fn=filter_fn
    )
    assert len(list(ds)) == num_docs + 1


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_doc_iter_method(filter_fn):
    ds = list(random_docs(10))

    for d in DocumentArray(ds):
        assert d.text == 'hello world'

    for d in DocumentArray(ds).traverse_flat('c,r', filter_fn=filter_fn):
        d.text = 'modified'

    for d in DocumentArray(ds):
        assert d.text == 'modified'


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_traverse_matcharray(filter_fn):
    doc = Document(
        matches=[
            Document(id=f'm{i}', chunks=[Document(id=f'm{i}c{j}') for j in range(3)])
            for i in range(3)
        ]
    )
    flat_docs = doc.matches.traverse_flat('r,c', filter_fn=filter_fn)
    assert isinstance(flat_docs, DocumentArray)
    assert len(flat_docs) == 12


@pytest.mark.parametrize('filter_fn', [(lambda d: True), None])
def test_traverse_chunkarray(filter_fn):
    doc = Document(
        chunks=[
            Document(id=f'c{i}', matches=[Document(id=f'c{i}m{j}') for j in range(3)])
            for i in range(3)
        ]
    )
    flat_docs = doc.chunks.traverse_flat('r,m', filter_fn=filter_fn)
    assert isinstance(flat_docs, DocumentArray)
    assert len(flat_docs) == 12


@pytest.mark.parametrize('use_dam', [True, False])
@pytest.mark.parametrize(
    ('filter_fn', 'docs_len'),
    [
        (lambda d: False, 0),
        (lambda d: d.text == 'hello', num_docs * num_matches_per_doc),
        (
            lambda d: d.text == 'world',
            num_docs * num_chunks_per_doc * num_matches_per_chunk,
        ),
        (
            lambda d: True,
            num_docs
            + num_docs * num_chunks_per_doc
            + num_docs * num_matches_per_doc
            + num_docs * num_chunks_per_doc * num_matches_per_chunk,
        ),
        (
            None,
            num_docs
            + num_docs * num_matches_per_doc
            + num_docs * num_chunks_per_doc
            + num_docs * num_chunks_per_doc * num_matches_per_chunk,
        ),
    ],
)
def test_filter_fn_traverse_flat(filter_fn, docs_len, doc_req, use_dam, tmp_path):
    docs = doc_req
    ds = list(docs.traverse_flat('r,c,m,cm', filter_fn=filter_fn))
    assert len(ds) == docs_len
    assert all(isinstance(d, Document) for d in ds)


@pytest.mark.parametrize('use_dam', [True, False])
@pytest.mark.parametrize(
    ('filter_fn', 'docs_len'),
    [
        (lambda d: False, [0, 0, 0, 0]),
        (lambda d: d.text == 'hello', [0, 0, num_docs * num_matches_per_doc, 0]),
        (
            lambda d: d.text == 'world',
            [0, 0, 0, num_docs * num_chunks_per_doc * num_matches_per_chunk],
        ),
        (
            lambda d: True,
            [
                num_docs,
                num_docs * num_chunks_per_doc,
                num_docs * num_matches_per_doc,
                num_docs * num_chunks_per_doc * num_matches_per_chunk,
            ],
        ),
        (
            None,
            [
                num_docs,
                num_docs * num_chunks_per_doc,
                num_docs * num_matches_per_doc,
                num_docs * num_chunks_per_doc * num_matches_per_chunk,
            ],
        ),
    ],
)
def test_filter_fn_traverse_flat_per_path(
    filter_fn, doc_req, docs_len, use_dam, tmp_path
):
    docs = doc_req
    ds = list(docs.traverse_flat_per_path('r,c,m,cm', filter_fn=filter_fn))
    assert len(ds) == 4
    for seq, length in zip(ds, docs_len):
        assert isinstance(seq, DocumentArray)
        assert len(list(seq)) == length


def test_traversal_path():
    da = DocumentArray([Document() for _ in range(6)])
    assert len(da) == 6

    da.traverse_flat('r')


def test_traverse_flat_root_itself():
    da = DocumentArray([Document() for _ in range(100)])
    res = da.traverse_flat('r')
    assert id(res) == id(da)


def da_and_dam(N):
    da = DocumentArray(random_docs(N))
    return (da,)


@pytest.mark.parametrize('da', da_and_dam(100))
def test_flatten(da):
    daf = da.flatten()
    assert len(daf) == 600
    assert isinstance(daf, DocumentArray)
    assert len(set(d.id for d in daf)) == 600

    # flattened DA can not be flattened again
    daf = daf.flatten()
    assert len(daf) == 600


def test_flatten_no_copy():
    da = da_and_dam(100)[0]
    daf = da.flatten()
    new_text = 'hi i changed it!'
    daf[53].text = new_text
    assert da[daf[53].id].text == new_text



import numpy as np
import pytest

from docarray import DocumentArray


@pytest.mark.parametrize('cls', [DocumentArray])
@pytest.mark.parametrize(
    'content_attr', ['texts', 'embeddings', 'blobs', 'buffers', 'contents']
)
def test_content_empty_getter_return_none(cls, content_attr):
    da = cls()
    assert getattr(da, content_attr) is None


@pytest.mark.parametrize('cls', [DocumentArray])
@pytest.mark.parametrize(
    'content_attr',
    [
        ('texts', ''),
        ('embeddings', np.array([])),
        ('blobs', np.array([])),
        ('buffers', []),
        ('contents', []),
    ],
)
def test_content_empty_setter(cls, content_attr):
    da = cls()
    setattr(da, content_attr[0], content_attr[1])
    assert getattr(da, content_attr[0]) is None


@pytest.mark.parametrize('cls', [DocumentArray])
@pytest.mark.parametrize(
    'content_attr',
    [
        ('texts', ['s'] * 10),
        ('blobs', np.random.random([10, 2])),
        ('buffers', [b's'] * 10),
    ],
)
def test_content_getter_setter(cls, content_attr):
    da = cls.empty(10)
    setattr(da, content_attr[0], content_attr[1])
    np.testing.assert_equal(da.contents, content_attr[1])
    da.contents = content_attr[1]
    np.testing.assert_equal(da.contents, content_attr[1])
    np.testing.assert_equal(getattr(da, content_attr[0]), content_attr[1])
    da.contents = None
    assert da.contents is None


@pytest.mark.parametrize('da_len', [0, 1, 2])
def test_content_empty(da_len):
    da = DocumentArray.empty(da_len)
    assert not da.texts
    assert not da.contents
    assert not da.blobs
    assert not da.buffers

    da.texts = ['hello'] * da_len
    if da_len == 0:
        assert not da.contents
    else:
        assert da.contents == ['hello'] * da_len
        assert da.texts == ['hello'] * da_len
        assert not da.blobs
        assert not da.buffers



import copy

import numpy as np
import pytest

from docarray import DocumentArray, Document


@pytest.mark.parametrize(
    'metric_fn, kwargs',
    [
        ('r_precision', {}),
        ('precision_at_k', {}),
        ('hit_at_k', {}),
        ('average_precision', {}),
        ('reciprocal_rank', {}),
        ('recall_at_k', {'max_rel': 9}),
        ('f1_score_at_k', {'max_rel': 9}),
        ('ndcg_at_k', {}),
    ],
)
def test_eval_mixin_perfect_match(metric_fn, kwargs):
    da = DocumentArray.empty(10)
    da.embeddings = np.random.random([10, 256])
    da.match(da, exclude_self=True)
    r = da.evaluate(da, metric=metric_fn, **kwargs)
    assert isinstance(r, float)
    assert r == 1.0
    for d in da:
        assert d.evaluations[metric_fn].value == 1.0


@pytest.mark.parametrize(
    'metric_fn, kwargs',
    [
        ('r_precision', {}),
        ('precision_at_k', {}),
        ('hit_at_k', {}),
        ('average_precision', {}),
        ('reciprocal_rank', {}),
        ('recall_at_k', {'max_rel': 9}),
        ('f1_score_at_k', {'max_rel': 9}),
        ('ndcg_at_k', {}),
    ],
)
def test_eval_mixin_zero_match(metric_fn, kwargs):
    da1 = DocumentArray.empty(10)
    da1.embeddings = np.random.random([10, 256])
    da1.match(da1, exclude_self=True)

    da2 = copy.deepcopy(da1)
    da2.embeddings = np.random.random([10, 256])
    da2.match(da2, exclude_self=True)

    r = da1.evaluate(da2, metric=metric_fn, **kwargs)
    assert isinstance(r, float)
    assert r == 1.0
    for d in da1:
        d: Document
        assert d.evaluations[metric_fn].value == 1.0


def test_diff_len_should_raise():
    da1 = DocumentArray.empty(10)
    da2 = DocumentArray.empty(5)
    with pytest.raises(ValueError):
        da1.evaluate(da2, metric='precision_at_k')


def test_diff_hash_fun_should_raise():
    da1 = DocumentArray.empty(10)
    da2 = DocumentArray.empty(10)
    with pytest.raises(ValueError):
        da1.evaluate(da2, metric='precision_at_k')


def test_same_hash_same_len_fun_should_work():
    da1 = DocumentArray.empty(10)
    da1.embeddings = np.random.random([10, 3])
    da1.match(da1)
    da2 = DocumentArray.empty(10)
    da2.embeddings = np.random.random([10, 3])
    da2.match(da2)
    with pytest.raises(ValueError):
        da1.evaluate(da2, metric='precision_at_k')
    for d1, d2 in zip(da1, da2):
        d1.id = d2.id

    da1.evaluate(da2, metric='precision_at_k')


def test_adding_noise():
    da = DocumentArray.empty(10)

    da.embeddings = np.random.random([10, 3])
    da.match(da, exclude_self=True)

    da2 = copy.deepcopy(da)

    for d in da2:
        d.matches.extend(DocumentArray.empty(10))
        d.matches = d.matches.shuffle()

    assert da2.evaluate(da, metric='precision_at_k', k=10) < 1.0

    for d in da2:
        assert 0.0 < d.evaluations['precision_at_k'].value < 1.0



from typing import (
    Union,
    TYPE_CHECKING,
    TypeVar,
    Sequence,
    Optional,
    List,
    Dict,
    Generator,
    Iterable,
    Tuple,
)

if TYPE_CHECKING:
    import scipy.sparse
    import tensorflow
    import torch
    import numpy as np

    from . import Document

    ArrayType = TypeVar(
        'ArrayType',
        np.ndarray,
        scipy.sparse.spmatrix,
        tensorflow.SparseTensor,
        tensorflow.Tensor,
        torch.Tensor,
        Sequence[float],
    )

    DocumentContentType = Union[bytes, str, ArrayType]
    ProtoValueType = Optional[Union[str, bool, float]]
    StructValueType = Union[
        ProtoValueType, List[ProtoValueType], Dict[str, ProtoValueType]
    ]

    DocumentArraySourceType = Union[
        Sequence[Document], Document, Generator[Document], Iterable[Document]
    ]
    T = TypeVar('T')

    AnyDNN = TypeVar(
        'AnyDNN'
    )  #: The type of any implementation of a Deep Neural Network object

    DocumentArraySingletonIndexType = Union[int, str]
    DocumentArrayMultipleIndexType = Union[
        slice, Sequence[int], Sequence[str], Sequence[bool], Ellipsis
    ]
    DocumentArraySingleAttributeType = Tuple[slice, str]
    DocumentArrayMultipleAttributeType = Tuple[slice, Sequence[str]]
    DocumentArrayIndexType = Union[
        DocumentArraySingletonIndexType,
        DocumentArrayMultipleIndexType,
        DocumentArraySingleAttributeType,
        DocumentArrayMultipleAttributeType,
    ]



import warnings
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ...types import T, AnyDNN


class EmbedMixin:
    """Helper functions for embedding with a model"""

    def embed(
        self: 'T',
        embed_model: 'AnyDNN',
        device: str = 'cpu',
        batch_size: int = 256,
        to_numpy: bool = False,
    ) -> 'T':
        """Fill :attr:`.embedding` of Documents inplace by using `embed_model`

        :param embed_model: the embedding model written in Keras/Pytorch/Paddle
        :param device: the computational device for `embed_model`, can be either
            `cpu` or `cuda`.
        :param batch_size: number of Documents in a batch for embedding
        :param to_numpy: if to store embeddings back to Document in ``numpy.ndarray`` or original framework format.
        :return: itself after modified.
        """

        fm = get_framework(embed_model)
        getattr(self, f'_set_embeddings_{fm}')(
            embed_model, device, batch_size, to_numpy
        )
        return self

    def _set_embeddings_keras(
        self: 'T',
        embed_model: 'AnyDNN',
        device: str = 'cpu',
        batch_size: int = 256,
        to_numpy: bool = False,
    ):
        import tensorflow as tf

        device = tf.device('/GPU:0') if device == 'cuda' else tf.device('/CPU:0')
        with device:
            for b in self.batch(batch_size):
                r = embed_model(b.blobs, training=False)
                b.embeddings = r.numpy() if to_numpy else r

    def _set_embeddings_torch(
        self: 'T',
        embed_model: 'AnyDNN',
        device: str = 'cpu',
        batch_size: int = 256,
        to_numpy: bool = False,
    ):
        import torch

        embed_model = embed_model.to(device)
        is_training_before = embed_model.training
        embed_model.eval()
        with torch.inference_mode():
            for b in self.batch(batch_size):
                batch_inputs = torch.tensor(b.blobs, device=device)
                r = embed_model(batch_inputs).cpu().detach()
                b.embeddings = r.numpy() if to_numpy else r
        if is_training_before:
            embed_model.train()

    def _set_embeddings_paddle(
        self: 'T',
        embed_model,
        device: str = 'cpu',
        batch_size: int = 256,
        to_numpy: bool = False,
    ):
        import paddle

        is_training_before = embed_model.training
        embed_model.to(device=device)
        embed_model.eval()
        for b in self.batch(batch_size):
            batch_inputs = paddle.to_tensor(b.blobs, place=device)
            r = embed_model(batch_inputs)
            b.embeddings = r.numpy() if to_numpy else r
        if is_training_before:
            embed_model.train()

    def _set_embeddings_onnx(
        self: 'T',
        embed_model,
        device: str = 'cpu',
        batch_size: int = 256,
        *args,
        **kwargs,
    ):
        # embed_model is always an onnx.InferenceSession
        if device != 'cpu':
            import onnxruntime as ort

            support_device = ort.get_device()
            if device.lower().strip() != support_device.lower().strip():
                warnings.warn(
                    f'Your installed `onnxruntime` supports `{support_device}`, but you give {device}'
                )

        for b in self.batch(batch_size):
            b.embeddings = embed_model.run(
                None, {embed_model.get_inputs()[0].name: b.blobs}
            )[0]


def get_framework(dnn_model) -> str:
    """Return the framework that powers a DNN model.

    .. note::
        This is not a solid implementation. It is based on ``__module__`` name,
        the key idea is to tell ``dnn_model`` without actually importing the
        framework.

    :param dnn_model: a DNN model
    :return: `keras`, `torch`, `paddle` or ValueError

    """
    import importlib.util

    if importlib.util.find_spec('torch'):
        import torch

        if isinstance(dnn_model, torch.nn.Module):
            return 'torch'

    if importlib.util.find_spec('paddle'):
        import paddle

        if isinstance(dnn_model, paddle.nn.Layer):
            return 'paddle'

    if importlib.util.find_spec('tensorflow'):
        from tensorflow import keras

        if isinstance(dnn_model, keras.layers.Layer):
            return 'keras'

    if importlib.util.find_spec('onnx'):
        from onnxruntime import InferenceSession

        if isinstance(dnn_model, InferenceSession):
            return 'onnx'

    raise ValueError(f'can not determine the backend of {dnn_model!r}')



import pytest

from docarray import Document, DocumentArray
from tests import random_docs


@pytest.mark.parametrize('protocol', ['protobuf', 'pickle'])
@pytest.mark.parametrize('compress', ['lz4', 'bz2', 'lzma', 'zlib', 'gzip', None])
def test_to_from_bytes(protocol, compress):
    d = Document(embedding=[1, 2, 3, 4, 5], text='hello')
    bstr = d.to_bytes(protocol=protocol, compress=compress)
    print(protocol, compress, len(bstr))
    d2 = Document.from_bytes(bstr, protocol=protocol, compress=compress)
    assert d2.non_empty_fields == d.non_empty_fields


@pytest.mark.parametrize('target', [DocumentArray.empty(10), random_docs(10)])
def test_dict_json(target):
    for d in target:
        d1 = Document.from_dict(d.to_dict())
        d2 = Document.from_json(d.to_json())
        assert d1 == d2



import numpy as np
import paddle
import pytest
import tensorflow as tf

from docarray.math.ndarray import to_numpy_array
import torch
from scipy.sparse import csr_matrix, coo_matrix, bsr_matrix, csc_matrix

from docarray import DocumentArray, Document


def get_ndarrays_for_ravel():
    a = np.random.random([10, 3])
    a[a > 0.5] = 0
    return [
        (a, False),
        (torch.tensor(a), False),
        (tf.constant(a), False),
        (paddle.to_tensor(a), False),
        (torch.tensor(a).to_sparse(), True),
        # (tf.sparse.from_dense(a), True),
        (csr_matrix(a), True),
        (bsr_matrix(a), True),
        (coo_matrix(a), True),
        (csc_matrix(a), True),
    ]


@pytest.mark.parametrize('ndarray_val, is_sparse', get_ndarrays_for_ravel())
@pytest.mark.parametrize('attr', ['embeddings', 'blobs'])
def test_ravel_embeddings_blobs(ndarray_val, attr, is_sparse):
    da = DocumentArray.empty(10)
    setattr(da, attr, ndarray_val)
    ndav = getattr(da, attr)

    # test read/getter
    assert type(ndav) is type(ndarray_val)

    if is_sparse:
        if hasattr(ndav, 'todense'):
            ndav = (ndav.todense(),)
            ndarray_val = ndarray_val.todense()
        if hasattr(ndav, 'to_dense'):
            ndav = (ndav.to_dense(),)
            ndarray_val = ndarray_val.to_dense()
        if isinstance(ndav, tf.SparseTensor):
            ndav = tf.sparse.to_dense(ndav)
            ndarray_val = tf.sparse.to_dense(ndarray_val)

    if isinstance(ndav, tuple):
        ndav = ndav[0]
    if hasattr(ndav, 'numpy'):
        ndav = ndav.numpy()
        ndarray_val = ndarray_val.numpy()

    np.testing.assert_almost_equal(ndav, ndarray_val)


@pytest.mark.parametrize('sparse_cls', [csr_matrix, csc_matrix, bsr_matrix, coo_matrix])
def test_bsr_coo_unravel(sparse_cls):
    a = np.random.random([10, 72])
    a[a > 0.5] = 0

    da = DocumentArray.empty(10)
    for d, a_row in zip(da, a):
        d.embedding = sparse_cls(a_row)

    np.testing.assert_almost_equal(a, da.embeddings.todense())


def get_ndarrays():
    a = np.random.random([10, 3])
    a[a > 0.5] = 0
    return [
        (a, False),
        (torch.tensor(a), False),
        (tf.constant(a), False),
        (paddle.to_tensor(a), False),
        (torch.tensor(a).to_sparse(), True),
        (tf.sparse.from_dense(a), True),
        (csr_matrix(a), True),
        (bsr_matrix(a), True),
        (coo_matrix(a), True),
        (csc_matrix(a), True),
    ]


@pytest.mark.parametrize('ndarray_val, is_sparse', get_ndarrays())
@pytest.mark.parametrize('attr', ['embedding', 'blob'])
def test_ndarray_force_numpy(ndarray_val, attr, is_sparse):
    d = Document()
    setattr(d, attr, ndarray_val)
    ndav = to_numpy_array(getattr(d, attr))
    assert isinstance(ndav, np.ndarray)
    assert ndav.shape == (10, 3)



import pytest

from docarray import DocumentArray, Document


def da_for_batching():
    da = DocumentArray.empty(100)
    return (da,)


def docarray_for_split():
    da = DocumentArray()
    da.append(Document(tags={'category': 'c'}))
    da.append(Document(tags={'category': 'c'}))
    da.append(Document(tags={'category': 'b'}))
    da.append(Document(tags={'category': 'a'}))
    da.append(Document(tags={'category': 'a'}))
    return (da,)


def docarray_for_split_at_zero():
    da = DocumentArray()
    da.append(Document(tags={'category': 0.0}))
    da.append(Document(tags={'category': 0.0}))
    da.append(Document(tags={'category': 1.0}))
    da.append(Document(tags={'category': 2.0}))
    da.append(Document(tags={'category': 2.0}))
    return (da,)


def docarray_for_nest_split():
    da = DocumentArray()
    da.append(Document(tags={'nest': {'category': 'c'}}))
    da.append(Document(tags={'nest': {'category': 'c'}}))
    da.append(Document(tags={'nest': {'category': 'b'}}))
    da.append(Document(tags={'nest': {'category': 'a'}}))
    da.append(Document(tags={'nest': {'category': 'a'}}))
    return (da,)


@pytest.mark.parametrize('da', docarray_for_split())
def test_split(da):
    rv = da.split_by_tag('category')
    assert isinstance(rv, dict)
    assert sorted(list(rv.keys())) == ['a', 'b', 'c']
    # assure order is preserved c, b, a
    assert list(rv.keys()) == ['c', 'b', 'a']
    # original input c, c, b, a, a
    assert len(rv['c']) == 2
    assert len(rv['b']) == 1
    assert len(rv['a']) == 2
    rv = da.split_by_tag('random')
    assert not rv  # wrong tag returns empty dict


@pytest.mark.parametrize('da', docarray_for_split_at_zero())
def test_split_at_zero(da):
    rv = da.split_by_tag('category')
    assert isinstance(rv, dict)
    assert sorted(list(rv.keys())) == [0.0, 1.0, 2.0]


@pytest.mark.parametrize('da', docarray_for_nest_split())
def test_dunder_split(da):
    rv = da.split_by_tag('nest__category')
    assert isinstance(rv, dict)
    assert sorted(list(rv.keys())) == ['a', 'b', 'c']
    # assure order is preserved c, b, a
    assert list(rv.keys()) == ['c', 'b', 'a']
    # original input c, c, b, a, a
    assert len(rv['c']) == 2
    assert len(rv['b']) == 1
    assert len(rv['a']) == 2

    assert len(da.split_by_tag('nest__random')) == 1


@pytest.mark.parametrize('da', da_for_batching())
@pytest.mark.parametrize('batch_size', [1, 5, 100, 200])
@pytest.mark.parametrize('shuffle', [True, False])
def test_batching(da, batch_size, shuffle):
    all_ids = []
    for v in da.batch(batch_size=batch_size, shuffle=shuffle):
        assert len(v) <= batch_size
        all_ids.extend(v.get_attributes('id'))

    if shuffle:
        assert all_ids != da.get_attributes('id')
    else:
        assert all_ids == da.get_attributes('id')



import warnings
from typing import List, Union, Any

from ...helper import typename, dunder_get


class GetAttributesMixin:
    """Provide helper functions for :class:`Document` to allow advanced set and get attributes """

    def get_attributes(self, *fields: str) -> Union[Any, List[Any]]:
        """Bulk fetch Document fields and return a list of the values of these fields

        :param fields: the variable length values to extract from the document
        :return: a list with the attributes of this document ordered as the args
        """

        ret = []
        for k in fields:
            if '__' in k:
                value = dunder_get(self, k)
            else:
                value = getattr(self, k)

            ret.append(value)

        # unboxing if args is single
        if len(fields) == 1:
            ret = ret[0]

        return ret



import pickle

import pytest

from docarray import Document
from docarray.document.data import DocumentData
from docarray.base import BaseDCType
from tests import random_docs


@pytest.mark.parametrize('cls', BaseDCType.__subclasses__())
def test_pickle_dump_load(cls):
    r = pickle.loads(pickle.dumps(cls()))
    isinstance(r, cls)


def test_pickle_dump_load_real_doc():
    for d in random_docs(10):
        dr = pickle.loads(pickle.dumps(d))
        assert dr == d
        assert dr.embedding is not None
        assert len(dr.chunks) == len(d.chunks)


def test_pickle_rely_on_data_class_and_document_class():
    # TODO (Han): This is not really a designed behavior, but atm I see no harm
    #  of having it, and no real usecases that against it.

    d = Document()
    d.id = 'hello'
    setattr(d, 'foo', 'bar')
    assert getattr(d, 'foo') == 'bar'
    r_d = Document.from_bytes(d.to_bytes(protocol='pickle'))
    assert r_d.id == d.id
    assert getattr(r_d, 'foo') == 'bar'



import pytest

from docarray import DocumentArray


def da_and_dam(N):
    da = DocumentArray.empty(N)
    return (da,)


@pytest.mark.parametrize('da', da_and_dam(100))
def test_sample(da):
    sampled = da.sample(1)
    assert len(sampled) == 1
    sampled = da.sample(5)
    assert len(sampled) == 5
    assert isinstance(sampled, DocumentArray)
    with pytest.raises(ValueError):
        da.sample(101)  # can not sample with k greater than lenth of document array.


@pytest.mark.parametrize('da', da_and_dam(100))
def test_sample_with_seed(da):
    sampled_1 = da.sample(5, seed=1)
    sampled_2 = da.sample(5, seed=1)
    sampled_3 = da.sample(5, seed=2)
    assert len(sampled_1) == len(sampled_2) == len(sampled_3) == 5
    assert sampled_1 == sampled_2
    assert sampled_1 != sampled_3


@pytest.mark.parametrize('da', da_and_dam(100))
def test_shuffle(da):
    shuffled = da.shuffle()
    assert len(shuffled) == len(da)
    assert isinstance(shuffled, DocumentArray)
    ids_before_shuffle = [d.id for d in da]
    ids_after_shuffle = [d.id for d in shuffled]
    assert ids_before_shuffle != ids_after_shuffle
    assert sorted(ids_before_shuffle) == sorted(ids_after_shuffle)


@pytest.mark.parametrize('da', da_and_dam(100))
def test_shuffle_with_seed(da):
    shuffled_1 = da.shuffle(seed=1)
    shuffled_2 = da.shuffle(seed=1)
    shuffled_3 = da.shuffle(seed=2)
    assert len(shuffled_1) == len(shuffled_2) == len(shuffled_3) == len(da)
    assert shuffled_1 == shuffled_2
    assert shuffled_1 != shuffled_3



import os
import shutil

import numpy as np
import pytest

from docarray import Document, DocumentArray
from docarray.document.generators import (
    from_files,
    from_ndarray,
    from_lines,
    from_csv,
    from_huggingface_datasets,
)

cur_dir = os.path.dirname(os.path.abspath(__file__))


@pytest.fixture(scope='function')
def filepath(tmpdir):
    input_filepath = os.path.join(tmpdir, 'input_file.csv')
    with open(input_filepath, 'w') as input_file:
        input_file.writelines(["1\n", "2\n", "3\n"])
    return input_filepath


@pytest.fixture(scope='function')
def dataset_configs():
    config = {
        'adversarial': {
            'dataset_path': 'adversarial_qa',
            'name': 'adversarialQA',
            'split': 'test',
        },
        'tweet_eval': {
            'dataset_path': 'tweet_eval',
            'name': 'emoji',
            'split': 'train+test',
        },
    }
    return config


def test_input_lines_with_filepath(filepath):
    result = list(from_lines(filepath=filepath, size=2))
    assert len(result) == 2
    assert isinstance(result[0], Document)


def test_input_csv_from_file():
    with open(os.path.join(cur_dir, 'toydata/docs.csv')) as fp:
        result = list(from_csv(fp))
    assert len(result) == 2
    assert isinstance(result[0], Document)
    assert result[0].tags['source'] == 'testsrc'


def test_input_csv_from_lines():
    with open(os.path.join(cur_dir, 'toydata/docs.csv')) as fp:
        result = list(from_lines(fp, line_format='csv'))
    assert len(result) == 2
    assert isinstance(result[0], Document)
    assert result[0].tags['source'] == 'testsrc'


def test_input_csv_from_lines_field_resolver():
    with open(os.path.join(cur_dir, 'toydata/docs.csv')) as fp:
        result = list(
            from_lines(fp, line_format='csv', field_resolver={'question': 'text'})
        )
    assert len(result) == 2
    assert isinstance(result[0], Document)
    assert result[0].tags['source'] == 'testsrc'
    assert not result[0].uri
    assert result[0].text


@pytest.mark.parametrize(
    'da',
    [
        DocumentArray,
    ],
)
def test_input_csv_from_strings(da):
    result = da.from_csv(os.path.join(cur_dir, 'toydata/docs.csv'))
    assert len(result) == 2
    assert isinstance(result[0], Document)
    assert result[0].tags['source'] == 'testsrc'


def test_input_lines_with_empty_filepath_and_lines():
    with pytest.raises(ValueError):
        lines = from_lines(lines=None, filepath=None)
        for _ in lines:
            pass


def test_input_lines_with_jsonlines_docs():
    result = list(from_lines(filepath=os.path.join(cur_dir, 'toydata/docs.jsonlines')))
    assert len(result) == 2
    assert result[0].text == "a"
    assert result[1].text == "b"


@pytest.mark.parametrize(
    'size, sampling_rate',
    [
        (None, None),
        (1, None),
        (None, 0.5),
    ],
)
@pytest.mark.parametrize(
    'da',
    [
        DocumentArray,
    ],
)
def test_input_lines_with_jsonlines_file(size, sampling_rate, da):
    result = da.from_lines(
        filepath=os.path.join(cur_dir, 'toydata/docs.jsonlines'),
        size=size,
        sampling_rate=sampling_rate,
    )

    assert len(result) == size if size is not None else 2
    if sampling_rate is None:
        assert result[0].text == "a"
        if size is None:
            assert result[1].text == "b"


@pytest.mark.parametrize(
    'size, sampling_rate',
    [
        (None, None),
        (1, None),
        (None, 0.5),
    ],
)
def test_input_lines_with_jsonslines(size, sampling_rate):
    with open(os.path.join(cur_dir, 'toydata/docs.jsonlines')) as fp:
        lines = fp.readlines()
    result = list(
        from_lines(
            lines=lines, line_format='json', size=size, sampling_rate=sampling_rate
        )
    )
    assert len(result) == size if size is not None else 2
    if sampling_rate is None:
        assert result[0].text == "a"
        if size is None:
            assert result[1].text == "b"


def test_input_lines_with_jsonlines_docs_groundtruth():
    result = list(
        from_lines(filepath=os.path.join(cur_dir, 'toydata/docs_groundtruth.jsonlines'))
    )
    assert len(result) == 2
    assert result[0][0].text == "a"
    assert result[0][1].text == "b"
    assert result[1][0].text == "c"
    assert result[1][1].text == "d"


@pytest.mark.parametrize(
    'size, sampling_rate',
    [
        (None, None),
        (1, None),
        (None, 0.5),
    ],
)
def test_input_huggingface_datasets_from_path(dataset_configs, size, sampling_rate):
    result = list(
        from_huggingface_datasets(
            dataset_configs['adversarial']['dataset_path'],
            size=size,
            name=dataset_configs['adversarial']['name'],
            sampling_rate=sampling_rate,
            split=dataset_configs['adversarial']['split'],
        )
    )

    if size is not None:
        assert len(result) == size

    assert isinstance(result[0], Document)


def test_input_huggingface_datasets_with_tweet_dataset(dataset_configs):
    result = list(
        from_huggingface_datasets(
            dataset_configs['tweet_eval']['dataset_path'],
            name=dataset_configs['tweet_eval']['name'],
            split=dataset_configs['tweet_eval']['split'],
        )
    )
    assert isinstance(result[0], Document)
    assert result[0].text


@pytest.mark.parametrize(
    'da',
    [
        DocumentArray,
    ],
)
def test_input_huggingface_datasets_from_csv_file(dataset_configs, da):
    field_resolver = {'question': 'text'}
    result = da.from_huggingface_datasets(
        'csv',
        field_resolver=field_resolver,
        data_files=os.path.join(cur_dir, 'toydata/docs.csv'),
        split='train',
    )

    assert len(result) == 2
    assert isinstance(result[0], Document)
    assert result[0].text == 'What are the symptoms?'
    assert result[0].tags['source'] == 'testsrc'


@pytest.mark.parametrize(
    'da',
    [
        DocumentArray,
    ],
)
def test_input_huggingface_datasets_with_field_resolver(dataset_configs, da):
    field_resolver = {'question': 'text'}
    result = da.from_huggingface_datasets(
        dataset_configs['adversarial']['dataset_path'],
        field_resolver=field_resolver,
        name=dataset_configs['adversarial']['name'],
        split=dataset_configs['adversarial']['split'],
    )

    assert isinstance(result[0], Document)
    assert result[0].text
    assert 'title' in result[0].tags


def test_input_huggingface_datasets_with_filter_fields(dataset_configs):
    field_resolver = {'question': 'text'}
    result = list(
        from_huggingface_datasets(
            dataset_configs['adversarial']['dataset_path'],
            field_resolver=field_resolver,
            filter_fields=True,
            name=dataset_configs['adversarial']['name'],
            split=dataset_configs['adversarial']['split'],
        )
    )
    assert isinstance(result[0], Document)
    assert result[0].text
    assert not 'title' in result[0].tags


def test_input_huggingface_datasets_with_no_split(dataset_configs):
    with pytest.raises(ValueError):
        result = from_huggingface_datasets(
            dataset_configs['adversarial']['dataset_path'],
            name=dataset_configs['adversarial']['name'],
        )
        for _ in result:
            pass


def test_input_huggingface_datasets_with_filter_fields_and_no_resolver(dataset_configs):
    with pytest.raises(ValueError):
        result = from_huggingface_datasets(
            dataset_configs['adversarial']['dataset_path'],
            name=dataset_configs['adversarial']['name'],
            filter_fields=True,
        )
        for _ in result:
            pass


@pytest.mark.parametrize(
    'patterns, recursive, size, sampling_rate, read_mode',
    [
        ('*.*', True, None, None, None),
        ('*.*', False, None, None, None),
        ('*.*', True, 2, None, None),
        ('*.*', True, 2, None, 'rb'),
        ('*.*', True, None, 0.5, None),
    ],
)
def test_input_files(patterns, recursive, size, sampling_rate, read_mode):
    DocumentArray(
        from_files(
            patterns=patterns,
            recursive=recursive,
            size=size,
            sampling_rate=sampling_rate,
            read_mode=read_mode,
        )
    )


def test_from_files_with_uri():
    for d in from_files(patterns='*.*', to_dataturi=True, size=10):
        assert d.uri.startswith('data:')



import os

import numpy as np
import onnxruntime
import paddle
import pytest
import tensorflow as tf
import torch

from docarray import DocumentArray

random_embed_models = {
    'keras': lambda: tf.keras.Sequential(
        [tf.keras.layers.Dropout(0.5), tf.keras.layers.BatchNormalization()]
    ),
    'pytorch': lambda: torch.nn.Sequential(
        torch.nn.Dropout(0.5), torch.nn.BatchNorm1d(128)
    ),
    'paddle': lambda: paddle.nn.Sequential(
        paddle.nn.Dropout(0.5), paddle.nn.BatchNorm1D(128)
    ),
}
cur_dir = os.path.dirname(os.path.abspath(__file__))
torch.onnx.export(
    random_embed_models['pytorch'](),
    torch.rand(1, 128),
    os.path.join(cur_dir, 'test-net.onnx'),
    do_constant_folding=True,  # whether to execute constant folding for optimization
    input_names=['input'],  # the model's input names
    output_names=['output'],  # the model's output names
    dynamic_axes={
        'input': {0: 'batch_size'},  # variable length axes
        'output': {0: 'batch_size'},
    },
)

random_embed_models['onnx'] = lambda: onnxruntime.InferenceSession(
    os.path.join(cur_dir, 'test-net.onnx')
)


@pytest.mark.parametrize('framework', ['onnx', 'keras', 'pytorch', 'paddle'])
@pytest.mark.parametrize('da', [DocumentArray])
@pytest.mark.parametrize('N', [2, 1000])
@pytest.mark.parametrize('batch_size', [1, 256])
@pytest.mark.parametrize('to_numpy', [True, False])
def test_embedding_on_random_network(framework, da, N, batch_size, to_numpy):
    docs = da.empty(N)
    docs.blobs = np.random.random([N, 128]).astype(np.float32)
    embed_model = random_embed_models[framework]()
    docs.embed(embed_model, batch_size=batch_size, to_numpy=to_numpy)

    r = docs.embeddings
    if hasattr(r, 'numpy'):
        r = r.numpy()
    embed1 = r.copy()

    # reset
    docs.embeddings = np.random.random([N, 128]).astype(np.float32)

    # try it again, it should yield the same result
    docs.embed(embed_model, batch_size=batch_size, to_numpy=to_numpy)
    np.testing.assert_array_almost_equal(docs.embeddings, embed1)

    # reset
    docs.embeddings = np.random.random([N, 128]).astype(np.float32)

    # now do this one by one
    docs[: int(N / 2)].embed(embed_model, batch_size=batch_size, to_numpy=to_numpy)
    docs[-int(N / 2) :].embed(embed_model, batch_size=batch_size, to_numpy=to_numpy)
    np.testing.assert_array_almost_equal(docs.embeddings, embed1)



import os
import uuid

import numpy as np
import pytest

from docarray import DocumentArray
from tests import random_docs


def da_and_dam():
    da = DocumentArray(random_docs(100))
    return (da,)


@pytest.mark.slow
@pytest.mark.parametrize('method', ['json', 'binary'])
@pytest.mark.parametrize('da', da_and_dam())
def test_document_save_load(method, tmp_path, da):
    tmp_file = os.path.join(tmp_path, 'test')
    da.save(tmp_file, file_format=method)
    da_r = type(da).load(tmp_file, file_format=method)

    assert type(da) is type(da_r)
    assert len(da) == len(da_r)
    for d, d_r in zip(da, da_r):
        assert d.id == d_r.id
        np.testing.assert_equal(d.embedding, d_r.embedding)
        assert d.content == d_r.content


@pytest.mark.parametrize('flatten_tags', [True, False])
@pytest.mark.parametrize('da', da_and_dam())
def test_da_csv_write(flatten_tags, tmp_path, da):
    tmpfile = os.path.join(tmp_path, 'test.csv')
    da.save_csv(tmpfile, flatten_tags)
    with open(tmpfile) as fp:
        assert len([v for v in fp]) == len(da) + 1


@pytest.mark.parametrize('da', [DocumentArray])
def test_from_ndarray(da):
    _da = da.from_ndarray(np.random.random([10, 256]))
    assert len(_da) == 10


@pytest.mark.parametrize('da', [DocumentArray])
def test_from_files(da):
    assert len(da.from_files(patterns='*.*', to_dataturi=True, size=1)) == 1


cur_dir = os.path.dirname(os.path.abspath(__file__))


@pytest.mark.parametrize('da', [DocumentArray])
def test_from_ndjson(da):
    with open(os.path.join(cur_dir, 'docs.jsonlines')) as fp:
        _da = da.from_ndjson(fp)
        assert len(_da) == 2


@pytest.mark.parametrize('da_cls', [DocumentArray])
def test_from_to_pd_dataframe(da_cls):
    # simple
    assert len(da_cls.from_dataframe(da_cls.empty(2).to_dataframe())) == 2

    # more complicated
    da = da_cls.empty(2)
    da.embeddings = [[1, 2, 3], [4, 5, 6]]
    da.blobs = [[1, 2], [2, 1]]
    da[0].tags = {'hello': 'world'}
    da2 = da_cls.from_dataframe(da.to_dataframe())
    assert da2[0].tags == {'hello': 'world'}
    assert da2[1].tags == {}


@pytest.mark.parametrize('da_cls', [DocumentArray])
def test_from_to_bytes(da_cls):
    # simple
    assert len(da_cls.load_binary(bytes(da_cls.empty(2)))) == 2

    # more complicated
    da = da_cls.empty(2)
    da.embeddings = [[1, 2, 3], [4, 5, 6]]
    da.blobs = [[1, 2], [2, 1]]
    da[0].tags = {'hello': 'world'}
    da2 = da_cls.load_binary(bytes(da))
    assert da2.blobs == [[1, 2], [2, 1]]
    assert da2.embeddings == [[1, 2, 3], [4, 5, 6]]
    assert da2[0].tags == {'hello': 'world'}
    assert da2[1].tags == {}


@pytest.mark.parametrize('da_cls', [DocumentArray])
@pytest.mark.parametrize('show_progress', [True, False])
def test_push_pull_io(da_cls, show_progress):
    da1 = da_cls.empty(10)
    da1.embeddings = np.random.random([len(da1), 256])
    random_texts = [str(uuid.uuid1()) for _ in da1]
    da1.texts = random_texts

    da1.push('myda', show_progress=show_progress)

    da2 = da_cls.pull('myda', show_progress=show_progress)

    assert len(da1) == len(da2) == 10
    assert da1.texts == da2.texts == random_texts



import pytest

from docarray.score import NamedScore


@pytest.mark.parametrize(
    'init_args', [None, dict(value=123, description='hello'), NamedScore()]
)
@pytest.mark.parametrize('copy', [True, False])
def test_construct_ns(init_args, copy):
    NamedScore(init_args, copy)



import numpy as np
import pytest
import scipy.sparse
import tensorflow as tf
import torch
from scipy.sparse import csr_matrix

from docarray import DocumentArray, Document
from tests import random_docs

rand_array = np.random.random([10, 3])


def da_and_dam():
    rand_docs = random_docs(100)
    da = DocumentArray()
    da.extend(rand_docs)
    return (da,)


@pytest.mark.parametrize(
    'array',
    [
        rand_array,
        torch.Tensor(rand_array),
        tf.constant(rand_array),
        csr_matrix(rand_array),
    ],
)
def test_set_embeddings_multi_kind(array):
    da = DocumentArray([Document() for _ in range(10)])
    da.embeddings = array


@pytest.mark.parametrize('da', da_and_dam())
def test_da_get_embeddings(da):
    np.testing.assert_almost_equal(da.get_attributes('embedding'), da.embeddings)


@pytest.mark.parametrize('da', da_and_dam())
def test_embeddings_setter_da(da):
    emb = np.random.random((100, 128))
    da.embeddings = emb
    np.testing.assert_almost_equal(da.embeddings, emb)

    for x, doc in zip(emb, da):
        np.testing.assert_almost_equal(x, doc.embedding)

    da.embeddings = None
    if hasattr(da, 'flush'):
        da.flush()
    assert not da.embeddings


@pytest.mark.parametrize('da', da_and_dam())
def test_embeddings_wrong_len(da):
    embeddings = np.ones((2, 10))

    with pytest.raises(ValueError):
        da.embeddings = embeddings


@pytest.mark.parametrize('da', da_and_dam())
def test_blobs_getter_da(da):
    blobs = np.random.random((100, 10, 10))
    da.blobs = blobs
    assert len(da) == 100
    np.testing.assert_almost_equal(da.get_attributes('blob'), da.blobs)
    np.testing.assert_almost_equal(da.blobs, blobs)

    da.blobs = None
    if hasattr(da, 'flush'):
        da.flush()
    assert not da.blobs


@pytest.mark.parametrize('da', da_and_dam())
def test_texts_getter_da(da):
    assert len(da.texts) == 100
    assert da.texts == da.get_attributes('text')
    texts = ['text' for _ in range(100)]
    da.texts = texts
    assert da.texts == texts

    for x, doc in zip(texts, da):
        assert x == doc.text

    da.texts = None
    if hasattr(da, 'flush'):
        da.flush()

    # unfortunately protobuf does not distinguish None and '' on string
    # so non-set str field in Pb is ''
    assert not da.texts


@pytest.mark.parametrize('da', da_and_dam())
def test_texts_wrong_len(da):
    texts = ['hello']

    with pytest.raises(ValueError):
        da.texts = texts


@pytest.mark.parametrize('da', da_and_dam())
def test_blobs_wrong_len(da):
    blobs = np.ones((2, 10, 10))

    with pytest.raises(ValueError):
        da.blobs = blobs


@pytest.mark.parametrize('da', da_and_dam())
def test_buffers_getter_setter(da):
    with pytest.raises(ValueError):
        da.buffers = [b'cc', b'bb', b'aa', b'dd']

    da.buffers = [b'aa'] * len(da)
    assert da.buffers == [b'aa'] * len(da)

    da.buffers = None
    if hasattr(da, 'flush'):
        da.flush()

    # unfortunately protobuf does not distinguish None and '' on string
    # so non-set str field in Pb is ''
    assert not da.buffers


def test_zero_embeddings():
    a = np.zeros([10, 6])
    da = DocumentArray.empty(10)

    # all zero, dense
    da.embeddings = a
    np.testing.assert_almost_equal(da.embeddings, a)
    for d in da:
        assert d.embedding.shape == (6,)

    # all zero, sparse
    sp_a = scipy.sparse.coo_matrix(a)
    da.embeddings = sp_a
    np.testing.assert_almost_equal(da.embeddings.todense(), sp_a.todense())
    for d in da:
        # scipy sparse row-vector can only be a (1, m) not squeezible
        assert d.embedding.shape == (1, 6)

    # near zero, sparse
    a = np.random.random([10, 6])
    a[a > 0.1] = 0
    sp_a = scipy.sparse.coo_matrix(a)
    da.embeddings = sp_a
    np.testing.assert_almost_equal(da.embeddings.todense(), sp_a.todense())
    for d in da:
        # scipy sparse row-vector can only be a (1, m) not squeezible
        assert d.embedding.shape == (1, 6)



import numpy as np
import pytest

from docarray import Document, DocumentArray
from docarray.array.chunk import ChunkArray
from docarray.array.match import MatchArray
from docarray.score import NamedScore


@pytest.mark.parametrize('init_args', [None, dict(id=123), Document()])
@pytest.mark.parametrize('copy', [True, False])
def test_construct_doc(init_args, copy):
    Document(init_args, copy)


def test_doc_hash_identical():
    d1 = Document(text='hello')
    d2 = Document(text='hello')
    assert hash(d1) != hash(d2)
    assert d1 != d2
    d1.id = d2.id
    assert hash(d1) == hash(d2)
    assert d1 == d2


def test_doc_hash_complicate_content():
    d1 = Document(text='hello', embedding=np.array([1, 2, 3]), id=1)
    d2 = Document(text='hello', embedding=np.array([1, 2, 3]), id=1)
    assert d1 == d2
    assert hash(d1) == hash(d2)


def test_pop_field():
    d1 = Document(text='hello', embedding=np.array([1, 2, 3]), id=1)
    assert d1.non_empty_fields == ('id', 'mime_type', 'text', 'embedding')
    d1.pop('text')
    assert d1.non_empty_fields == ('id', 'mime_type', 'embedding')
    d1.pop('id', 'embedding', 'mime_type')
    assert d1.non_empty_fields == tuple()

    d1.pop('foobar')
    with pytest.raises(AttributeError):
        assert d1.foobar


def test_clear_fields():
    d1 = Document(text='hello', embedding=np.array([1, 2, 3]), id=1)
    d1.clear()
    assert d1.non_empty_fields == tuple()


def test_exclusive_content():
    d = Document(text='hello')
    assert d.content_type == 'text'
    d.buffer = b'123'
    assert d.buffer
    assert not d.text
    assert not d.blob
    assert d.content_type == 'buffer'
    d.blob = [1, 2, 3]
    assert d.blob
    assert not d.buffer
    assert not d.text
    assert d.content_type == 'blob'
    d.text = 'hello'
    assert d.text
    assert not d.buffer
    assert not d.blob
    assert d.content_type == 'text'


def test_content_setter():
    d = Document()
    assert not d.content_type
    d.content = 'hello'
    assert d.content_type == 'text'
    d.content = None
    assert not d.content_type


def test_chunks_matches_setter():
    d = Document(chunks=[Document()], matches=[Document(), Document()])
    assert len(d.chunks) == 1
    assert len(d.matches) == 2
    assert isinstance(d.chunks, DocumentArray)
    assert isinstance(d.chunks, ChunkArray)
    assert isinstance(d.matches, DocumentArray)
    assert isinstance(d.matches, MatchArray)


def test_empty_doc_chunks_matches():
    assert isinstance(Document().chunks, DocumentArray)
    assert isinstance(Document().matches, DocumentArray)
    assert isinstance(Document().matches, MatchArray)
    assert isinstance(Document().chunks, ChunkArray)

    d = Document()
    d.chunks.append(Document())
    assert isinstance(d.chunks, ChunkArray)

    d.chunks = [Document(), Document()]
    assert isinstance(d.chunks, ChunkArray)


def test_chunk_match_increase_granularity():
    d = Document()
    d.chunks.append(Document())
    assert d.chunks[0].granularity == 1
    assert id(d.chunks.reference_doc) == id(d)
    d.matches.append(Document())
    assert d.matches[0].adjacency == 1
    assert id(d.matches.reference_doc) == id(d)

    d = d.chunks[0]
    d.chunks.append(Document())
    assert d.chunks[0].granularity == 2
    assert id(d.chunks.reference_doc) == id(d)

    d.matches.append(Document())
    assert d.matches[0].adjacency == 1
    assert id(d.matches.reference_doc) == id(d)


def test_offset():
    d1 = Document(offset=1.0)
    d2 = Document()
    d2.offset = 1.0
    assert d1.offset == d2.offset == 1.0


def test_exclusive_content_2():
    d = Document(text='hello', buffer=b'sda')
    assert len(d.non_empty_fields) == 3
    d.content = b'sda'
    assert d.content == b'sda'
    assert 'buffer' in d.non_empty_fields
    d = Document(content='hello')
    assert d.content_type == 'text'
    d = Document(content=b'hello')
    assert d.content_type == 'buffer'
    d = Document(content=[1, 2, 3])
    assert d.content_type == 'blob'


def test_get_attr_values():
    d = Document(
        **{
            'id': '123',
            'text': 'document',
            'feature1': 121,
            'name': 'name',
            'tags': {'id': 'identity', 'a': 'b', 'c': 'd', 'e': [0, 1, {'f': 'g'}]},
        }
    )
    d.scores['metric'] = NamedScore(value=42)

    required_keys = [
        'id',
        'text',
        'tags__name',
        'tags__feature1',
        'scores__metric__value',
        'tags__c',
        'tags__id',
        'tags__e__2__f',
    ]
    res = d.get_attributes(*required_keys)
    assert len(res) == len(required_keys)
    assert res[required_keys.index('id')] == '123'
    assert res[required_keys.index('tags__feature1')] == 121
    assert res[required_keys.index('tags__name')] == 'name'
    assert res[required_keys.index('text')] == 'document'
    assert res[required_keys.index('tags__c')] == 'd'
    assert res[required_keys.index('tags__id')] == 'identity'
    assert res[required_keys.index('scores__metric__value')] == 42
    assert res[required_keys.index('tags__e__2__f')] == 'g'

    required_keys_2 = ['tags', 'text']
    res2 = d.get_attributes(*required_keys_2)
    assert len(res2) == 2
    assert res2[required_keys_2.index('text')] == 'document'
    assert res2[required_keys_2.index('tags')] == d.tags

    d = Document({'id': '123', 'tags': {'outterkey': {'innerkey': 'real_value'}}})
    required_keys_3 = ['tags__outterkey__innerkey']
    res3 = d.get_attributes(*required_keys_3)
    assert res3 == 'real_value'

    d = Document(content=np.array([1, 2, 3]))
    res4 = np.stack(d.get_attributes(*['blob']))
    np.testing.assert_equal(res4, np.array([1, 2, 3]))


def test_set_get_mime():
    a = Document()
    a.mime_type = 'jpg'
    assert a.mime_type == 'image/jpeg'
    b = Document()
    b.mime_type = 'jpeg'
    assert b.mime_type == 'image/jpeg'
    c = Document()
    c.mime_type = '.jpg'
    assert c.mime_type == 'image/jpeg'


def test_doc_content():
    d = Document()
    assert d.content is None
    d.text = 'abc'
    assert d.content == 'abc'
    c = np.random.random([10, 10])
    d.blob = c
    np.testing.assert_equal(d.content, c)
    d.buffer = b'123'
    assert d.buffer == b'123'


def test_dict_constructor():

    d1 = Document(
        uri='https://jina.ai', mime_type='text/plain', granularity=1, adjacency=3
    )

    d2 = Document(
        dict(uri='https://jina.ai', mime_type='text/plain', granularity=1, adjacency=3)
    )

    d3 = Document(
        {
            'uri': 'https://jina.ai',
            'mime_type': 'text/plain',
            'granularity': 1,
            'adjacency': 3,
        }
    )

    assert d1 != d2
    d1.id = None
    d2.id = None
    d3.id = None
    assert d1 == d2 == d3


def test_unknown_fields_behavior():
    d = Document(hello='world')
    assert d.tags == {'hello': 'world'}

    d = Document(hello='world', unknown_fields_handler='drop')
    assert d.tags == {}

    with pytest.raises(AttributeError):
        d = Document(hello='world', unknown_fields_handler='raise')


def test_content_setter_as_proxy():
    d = Document(content='hello')
    assert d.content == 'hello'

    assert 'content' not in d.non_empty_fields
    assert 'text' in d.non_empty_fields
    d.content = [1, 2, 3]
    assert 'blob' in d.non_empty_fields
    assert 'text' not in d.non_empty_fields



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: docarray.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='docarray.proto',
  package='docarray',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x0e\x64ocarray.proto\x12\x08\x64ocarray\x1a\x1cgoogle/protobuf/struct.proto\"A\n\x11\x44\x65nseNdArrayProto\x12\x0e\n\x06\x62uffer\x18\x01 \x01(\x0c\x12\r\n\x05shape\x18\x02 \x03(\r\x12\r\n\x05\x64type\x18\x03 \x01(\t\"\xb6\x01\n\x0cNdArrayProto\x12,\n\x05\x64\x65nse\x18\x01 \x01(\x0b\x32\x1b.docarray.DenseNdArrayProtoH\x00\x12.\n\x06sparse\x18\x02 \x01(\x0b\x32\x1c.docarray.SparseNdArrayProtoH\x00\x12\x10\n\x08\x63ls_name\x18\x03 \x01(\t\x12+\n\nparameters\x18\x04 \x01(\x0b\x32\x17.google.protobuf.StructB\t\n\x07\x63ontent\"~\n\x12SparseNdArrayProto\x12,\n\x07indices\x18\x01 \x01(\x0b\x32\x1b.docarray.DenseNdArrayProto\x12+\n\x06values\x18\x02 \x01(\x0b\x32\x1b.docarray.DenseNdArrayProto\x12\r\n\x05shape\x18\x03 \x03(\r\"V\n\x0fNamedScoreProto\x12\r\n\x05value\x18\x01 \x01(\x02\x12\x0f\n\x07op_name\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x0e\n\x06ref_id\x18\x04 \x01(\t\"\xc1\x05\n\rDocumentProto\x12\n\n\x02id\x18\x01 \x01(\t\x12\x10\n\x06\x62uffer\x18\x02 \x01(\x0cH\x00\x12&\n\x04\x62lob\x18\x03 \x01(\x0b\x32\x16.docarray.NdArrayProtoH\x00\x12\x0e\n\x04text\x18\x04 \x01(\tH\x00\x12\x13\n\x0bgranularity\x18\x05 \x01(\r\x12\x11\n\tadjacency\x18\x06 \x01(\r\x12\x11\n\tparent_id\x18\x07 \x01(\t\x12\x0e\n\x06weight\x18\x08 \x01(\x02\x12\x0b\n\x03uri\x18\t \x01(\t\x12\x10\n\x08modality\x18\n \x01(\t\x12\x11\n\tmime_type\x18\x0b \x01(\t\x12\x0e\n\x06offset\x18\x0c \x01(\x02\x12\x10\n\x08location\x18\r \x03(\x02\x12\'\n\x06\x63hunks\x18\x0e \x03(\x0b\x32\x17.docarray.DocumentProto\x12(\n\x07matches\x18\x0f \x03(\x0b\x32\x17.docarray.DocumentProto\x12)\n\tembedding\x18\x10 \x01(\x0b\x32\x16.docarray.NdArrayProto\x12%\n\x04tags\x18\x11 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x33\n\x06scores\x18\x12 \x03(\x0b\x32#.docarray.DocumentProto.ScoresEntry\x12=\n\x0b\x65valuations\x18\x13 \x03(\x0b\x32(.docarray.DocumentProto.EvaluationsEntry\x1aH\n\x0bScoresEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.docarray.NamedScoreProto:\x02\x38\x01\x1aM\n\x10\x45valuationsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.docarray.NamedScoreProto:\x02\x38\x01\x42\t\n\x07\x63ontent\";\n\x12\x44ocumentArrayProto\x12%\n\x04\x64ocs\x18\x01 \x03(\x0b\x32\x17.docarray.DocumentProtob\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,])




_DENSENDARRAYPROTO = _descriptor.Descriptor(
  name='DenseNdArrayProto',
  full_name='docarray.DenseNdArrayProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='buffer', full_name='docarray.DenseNdArrayProto.buffer', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='shape', full_name='docarray.DenseNdArrayProto.shape', index=1,
      number=2, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dtype', full_name='docarray.DenseNdArrayProto.dtype', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=58,
  serialized_end=123,
)


_NDARRAYPROTO = _descriptor.Descriptor(
  name='NdArrayProto',
  full_name='docarray.NdArrayProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='dense', full_name='docarray.NdArrayProto.dense', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sparse', full_name='docarray.NdArrayProto.sparse', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cls_name', full_name='docarray.NdArrayProto.cls_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parameters', full_name='docarray.NdArrayProto.parameters', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='content', full_name='docarray.NdArrayProto.content',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=126,
  serialized_end=308,
)


_SPARSENDARRAYPROTO = _descriptor.Descriptor(
  name='SparseNdArrayProto',
  full_name='docarray.SparseNdArrayProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='indices', full_name='docarray.SparseNdArrayProto.indices', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='values', full_name='docarray.SparseNdArrayProto.values', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='shape', full_name='docarray.SparseNdArrayProto.shape', index=2,
      number=3, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=310,
  serialized_end=436,
)


_NAMEDSCOREPROTO = _descriptor.Descriptor(
  name='NamedScoreProto',
  full_name='docarray.NamedScoreProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='docarray.NamedScoreProto.value', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='op_name', full_name='docarray.NamedScoreProto.op_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='docarray.NamedScoreProto.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ref_id', full_name='docarray.NamedScoreProto.ref_id', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=438,
  serialized_end=524,
)


_DOCUMENTPROTO_SCORESENTRY = _descriptor.Descriptor(
  name='ScoresEntry',
  full_name='docarray.DocumentProto.ScoresEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='docarray.DocumentProto.ScoresEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='docarray.DocumentProto.ScoresEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1070,
  serialized_end=1142,
)

_DOCUMENTPROTO_EVALUATIONSENTRY = _descriptor.Descriptor(
  name='EvaluationsEntry',
  full_name='docarray.DocumentProto.EvaluationsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='docarray.DocumentProto.EvaluationsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='docarray.DocumentProto.EvaluationsEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1144,
  serialized_end=1221,
)

_DOCUMENTPROTO = _descriptor.Descriptor(
  name='DocumentProto',
  full_name='docarray.DocumentProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='docarray.DocumentProto.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='buffer', full_name='docarray.DocumentProto.buffer', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='blob', full_name='docarray.DocumentProto.blob', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='text', full_name='docarray.DocumentProto.text', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='granularity', full_name='docarray.DocumentProto.granularity', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='adjacency', full_name='docarray.DocumentProto.adjacency', index=5,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parent_id', full_name='docarray.DocumentProto.parent_id', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='weight', full_name='docarray.DocumentProto.weight', index=7,
      number=8, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='uri', full_name='docarray.DocumentProto.uri', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='modality', full_name='docarray.DocumentProto.modality', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mime_type', full_name='docarray.DocumentProto.mime_type', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='offset', full_name='docarray.DocumentProto.offset', index=11,
      number=12, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='location', full_name='docarray.DocumentProto.location', index=12,
      number=13, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='chunks', full_name='docarray.DocumentProto.chunks', index=13,
      number=14, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='matches', full_name='docarray.DocumentProto.matches', index=14,
      number=15, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='embedding', full_name='docarray.DocumentProto.embedding', index=15,
      number=16, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tags', full_name='docarray.DocumentProto.tags', index=16,
      number=17, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='scores', full_name='docarray.DocumentProto.scores', index=17,
      number=18, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='evaluations', full_name='docarray.DocumentProto.evaluations', index=18,
      number=19, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_DOCUMENTPROTO_SCORESENTRY, _DOCUMENTPROTO_EVALUATIONSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='content', full_name='docarray.DocumentProto.content',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=527,
  serialized_end=1232,
)


_DOCUMENTARRAYPROTO = _descriptor.Descriptor(
  name='DocumentArrayProto',
  full_name='docarray.DocumentArrayProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='docs', full_name='docarray.DocumentArrayProto.docs', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1234,
  serialized_end=1293,
)

_NDARRAYPROTO.fields_by_name['dense'].message_type = _DENSENDARRAYPROTO
_NDARRAYPROTO.fields_by_name['sparse'].message_type = _SPARSENDARRAYPROTO
_NDARRAYPROTO.fields_by_name['parameters'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_NDARRAYPROTO.oneofs_by_name['content'].fields.append(
  _NDARRAYPROTO.fields_by_name['dense'])
_NDARRAYPROTO.fields_by_name['dense'].containing_oneof = _NDARRAYPROTO.oneofs_by_name['content']
_NDARRAYPROTO.oneofs_by_name['content'].fields.append(
  _NDARRAYPROTO.fields_by_name['sparse'])
_NDARRAYPROTO.fields_by_name['sparse'].containing_oneof = _NDARRAYPROTO.oneofs_by_name['content']
_SPARSENDARRAYPROTO.fields_by_name['indices'].message_type = _DENSENDARRAYPROTO
_SPARSENDARRAYPROTO.fields_by_name['values'].message_type = _DENSENDARRAYPROTO
_DOCUMENTPROTO_SCORESENTRY.fields_by_name['value'].message_type = _NAMEDSCOREPROTO
_DOCUMENTPROTO_SCORESENTRY.containing_type = _DOCUMENTPROTO
_DOCUMENTPROTO_EVALUATIONSENTRY.fields_by_name['value'].message_type = _NAMEDSCOREPROTO
_DOCUMENTPROTO_EVALUATIONSENTRY.containing_type = _DOCUMENTPROTO
_DOCUMENTPROTO.fields_by_name['blob'].message_type = _NDARRAYPROTO
_DOCUMENTPROTO.fields_by_name['chunks'].message_type = _DOCUMENTPROTO
_DOCUMENTPROTO.fields_by_name['matches'].message_type = _DOCUMENTPROTO
_DOCUMENTPROTO.fields_by_name['embedding'].message_type = _NDARRAYPROTO
_DOCUMENTPROTO.fields_by_name['tags'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_DOCUMENTPROTO.fields_by_name['scores'].message_type = _DOCUMENTPROTO_SCORESENTRY
_DOCUMENTPROTO.fields_by_name['evaluations'].message_type = _DOCUMENTPROTO_EVALUATIONSENTRY
_DOCUMENTPROTO.oneofs_by_name['content'].fields.append(
  _DOCUMENTPROTO.fields_by_name['buffer'])
_DOCUMENTPROTO.fields_by_name['buffer'].containing_oneof = _DOCUMENTPROTO.oneofs_by_name['content']
_DOCUMENTPROTO.oneofs_by_name['content'].fields.append(
  _DOCUMENTPROTO.fields_by_name['blob'])
_DOCUMENTPROTO.fields_by_name['blob'].containing_oneof = _DOCUMENTPROTO.oneofs_by_name['content']
_DOCUMENTPROTO.oneofs_by_name['content'].fields.append(
  _DOCUMENTPROTO.fields_by_name['text'])
_DOCUMENTPROTO.fields_by_name['text'].containing_oneof = _DOCUMENTPROTO.oneofs_by_name['content']
_DOCUMENTARRAYPROTO.fields_by_name['docs'].message_type = _DOCUMENTPROTO
DESCRIPTOR.message_types_by_name['DenseNdArrayProto'] = _DENSENDARRAYPROTO
DESCRIPTOR.message_types_by_name['NdArrayProto'] = _NDARRAYPROTO
DESCRIPTOR.message_types_by_name['SparseNdArrayProto'] = _SPARSENDARRAYPROTO
DESCRIPTOR.message_types_by_name['NamedScoreProto'] = _NAMEDSCOREPROTO
DESCRIPTOR.message_types_by_name['DocumentProto'] = _DOCUMENTPROTO
DESCRIPTOR.message_types_by_name['DocumentArrayProto'] = _DOCUMENTARRAYPROTO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DenseNdArrayProto = _reflection.GeneratedProtocolMessageType('DenseNdArrayProto', (_message.Message,), {
  'DESCRIPTOR' : _DENSENDARRAYPROTO,
  '__module__' : 'docarray_pb2'
  # @@protoc_insertion_point(class_scope:docarray.DenseNdArrayProto)
  })
_sym_db.RegisterMessage(DenseNdArrayProto)

NdArrayProto = _reflection.GeneratedProtocolMessageType('NdArrayProto', (_message.Message,), {
  'DESCRIPTOR' : _NDARRAYPROTO,
  '__module__' : 'docarray_pb2'
  # @@protoc_insertion_point(class_scope:docarray.NdArrayProto)
  })
_sym_db.RegisterMessage(NdArrayProto)

SparseNdArrayProto = _reflection.GeneratedProtocolMessageType('SparseNdArrayProto', (_message.Message,), {
  'DESCRIPTOR' : _SPARSENDARRAYPROTO,
  '__module__' : 'docarray_pb2'
  # @@protoc_insertion_point(class_scope:docarray.SparseNdArrayProto)
  })
_sym_db.RegisterMessage(SparseNdArrayProto)

NamedScoreProto = _reflection.GeneratedProtocolMessageType('NamedScoreProto', (_message.Message,), {
  'DESCRIPTOR' : _NAMEDSCOREPROTO,
  '__module__' : 'docarray_pb2'
  # @@protoc_insertion_point(class_scope:docarray.NamedScoreProto)
  })
_sym_db.RegisterMessage(NamedScoreProto)

DocumentProto = _reflection.GeneratedProtocolMessageType('DocumentProto', (_message.Message,), {

  'ScoresEntry' : _reflection.GeneratedProtocolMessageType('ScoresEntry', (_message.Message,), {
    'DESCRIPTOR' : _DOCUMENTPROTO_SCORESENTRY,
    '__module__' : 'docarray_pb2'
    # @@protoc_insertion_point(class_scope:docarray.DocumentProto.ScoresEntry)
    })
  ,

  'EvaluationsEntry' : _reflection.GeneratedProtocolMessageType('EvaluationsEntry', (_message.Message,), {
    'DESCRIPTOR' : _DOCUMENTPROTO_EVALUATIONSENTRY,
    '__module__' : 'docarray_pb2'
    # @@protoc_insertion_point(class_scope:docarray.DocumentProto.EvaluationsEntry)
    })
  ,
  'DESCRIPTOR' : _DOCUMENTPROTO,
  '__module__' : 'docarray_pb2'
  # @@protoc_insertion_point(class_scope:docarray.DocumentProto)
  })
_sym_db.RegisterMessage(DocumentProto)
_sym_db.RegisterMessage(DocumentProto.ScoresEntry)
_sym_db.RegisterMessage(DocumentProto.EvaluationsEntry)

DocumentArrayProto = _reflection.GeneratedProtocolMessageType('DocumentArrayProto', (_message.Message,), {
  'DESCRIPTOR' : _DOCUMENTARRAYPROTO,
  '__module__' : 'docarray_pb2'
  # @@protoc_insertion_point(class_scope:docarray.DocumentArrayProto)
  })
_sym_db.RegisterMessage(DocumentArrayProto)


_DOCUMENTPROTO_SCORESENTRY._options = None
_DOCUMENTPROTO_EVALUATIONSENTRY._options = None
# @@protoc_insertion_point(module_scope)



import pytest

from docarray import Document
from docarray.array.chunk import ChunkArray
from docarray.array.match import MatchArray
from docarray.score import NamedScore


@pytest.mark.parametrize(
    'obj',
    [
        Document(),
        NamedScore(),
        MatchArray([Document()], Document()),
        ChunkArray([Document()], Document()),
    ],
)
def test_builtin_str_repr_no_content(obj):
    print(obj)
    print(f'{obj!r}')


@pytest.mark.parametrize(
    'obj',
    [
        Document(content='123', chunks=[Document(content='abc')]),
        NamedScore(
            op_name='operation',
            value=10.0,
            ref_id='10' * 16,
            description='score description',
        ),
    ],
)
def test_builtin_str_repr_has_content(obj):
    print(obj)
    print(f'{obj!r}')



import os

from docarray import Document

cur_dir = os.path.dirname(os.path.abspath(__file__))


def test_single_doc_summary():
    # empty doc
    Document().summary()
    # nested doc
    Document(
        chunks=[
            Document(),
            Document(chunks=[Document()]),
            Document(),
        ],
        matches=[Document(), Document()],
    ).summary()


def test_plot_image():
    d = Document(uri=os.path.join(cur_dir, 'toydata/test.png'))
    d.plot()

    d.load_uri_to_image_blob()
    d.uri = None

    d.plot()



import numpy as np
import pytest

from docarray import DocumentArray, Document


def da_and_dam():
    da = DocumentArray(
        [
            Document(text='hello'),
            Document(text='hello world'),
            Document(text='goodbye world!'),
        ]
    )

    return (da,)


@pytest.mark.parametrize('min_freq', [1, 2, 3])
@pytest.mark.parametrize('da', da_and_dam())
def test_da_vocabulary(da, min_freq):
    vocab = da.get_vocabulary(min_freq)
    if min_freq <= 1:
        assert set(vocab.values()) == {2, 3, 4}  # 0,1 are reserved
        assert set(vocab.keys()) == {'hello', 'world', 'goodbye'}
    elif min_freq == 2:
        assert set(vocab.values()) == {2, 3}  # 0,1 are reserved
        assert set(vocab.keys()) == {'hello', 'world'}
    elif min_freq == 3:
        assert not vocab.values()
        assert not vocab.keys()


@pytest.mark.parametrize('test_docs', da_and_dam())
def test_da_text_to_blob_non_max_len(test_docs):
    vocab = test_docs.get_vocabulary()
    for d in test_docs:
        d.convert_text_to_blob(vocab)
    np.testing.assert_array_equal(test_docs[0].blob, [2])
    np.testing.assert_array_equal(test_docs[1].blob, [2, 3])
    np.testing.assert_array_equal(test_docs[2].blob, [4, 3])
    for d in test_docs:
        d.convert_blob_to_text(vocab)

    assert test_docs[0].text == 'hello'
    assert test_docs[1].text == 'hello world'
    assert test_docs[2].text == 'goodbye world'


@pytest.mark.parametrize('test_docs', da_and_dam())
def test_da_text_to_blob_max_len_3(test_docs):
    vocab = test_docs.get_vocabulary()
    for d in test_docs:
        d.convert_text_to_blob(vocab, max_length=3)
    np.testing.assert_array_equal(test_docs[0].blob, [0, 0, 2])
    np.testing.assert_array_equal(test_docs[1].blob, [0, 2, 3])
    np.testing.assert_array_equal(test_docs[2].blob, [0, 4, 3])
    for d in test_docs:
        d.convert_blob_to_text(vocab)

    assert test_docs[0].text == 'hello'
    assert test_docs[1].text == 'hello world'
    assert test_docs[2].text == 'goodbye world'


@pytest.mark.parametrize('test_docs', da_and_dam())
def test_da_text_to_blob_max_len_1(test_docs):
    vocab = test_docs.get_vocabulary()
    for d in test_docs:
        d.convert_text_to_blob(vocab, max_length=1)
    np.testing.assert_array_equal(test_docs[0].blob, [2])
    np.testing.assert_array_equal(test_docs[1].blob, [3])
    np.testing.assert_array_equal(test_docs[2].blob, [3])
    for d in test_docs:
        d.convert_blob_to_text(vocab)

    assert test_docs[0].text == 'hello'
    assert test_docs[1].text == 'world'
    assert test_docs[2].text == 'world'


@pytest.mark.parametrize('da', da_and_dam())
def test_convert_text_blob_random_text(da):
    texts = ['a short phrase', 'word', 'this is a much longer sentence']
    da.clear()
    da.extend(Document(text=t) for t in texts)
    vocab = da.get_vocabulary()

    # encoding
    for d in da:
        d.convert_text_to_blob(vocab, max_length=10)

    # decoding
    for d in da:
        d.convert_blob_to_text(vocab)

    assert texts
    assert da.texts == texts



import csv
import glob
import itertools
import json
import os
import random
from contextlib import nullcontext
from typing import (
    Optional,
    Generator,
    Union,
    List,
    Iterable,
    Dict,
    TYPE_CHECKING,
    TextIO,
)

import numpy as np

if TYPE_CHECKING:
    from .. import Document


def from_ndarray(
    array: 'np.ndarray',
    axis: int = 0,
    size: Optional[int] = None,
    shuffle: bool = False,
) -> Generator['Document', None, None]:
    """Create a generator for a given dimension of a numpy array.

    :param array: the numpy ndarray data source
    :param axis: iterate over that axis
    :param size: the maximum number of the sub arrays
    :param shuffle: shuffle the numpy data source beforehand
    :yield: documents
    """

    from . import Document

    if shuffle:
        # shuffle for random query
        array = np.take(array, np.random.permutation(array.shape[0]), axis=axis)
    d = 0
    for r in array:
        yield Document(content=r)
        d += 1
        if size is not None and d >= size:
            break


def from_files(
    patterns: Union[str, List[str]],
    recursive: bool = True,
    size: Optional[int] = None,
    sampling_rate: Optional[float] = None,
    read_mode: Optional[str] = None,
    to_dataturi: bool = False,
) -> Generator['Document', None, None]:
    """Creates an iterator over a list of file path or the content of the files.

    :param patterns: The pattern may contain simple shell-style wildcards, e.g. '\*.py', '[\*.zip, \*.gz]'
    :param recursive: If recursive is true, the pattern '**' will match any files
        and zero or more directories and subdirectories
    :param size: the maximum number of the files
    :param sampling_rate: the sampling rate between [0, 1]
    :param read_mode: specifies the mode in which the file is opened.
        'r' for reading in text mode, 'rb' for reading in binary mode.
        If `read_mode` is None, will iterate over filenames.
    :param to_dataturi: if set, then the Document.uri will be filled with DataURI instead of the plan URI
    :yield: file paths or binary content

    .. note::
        This function should not be directly used, use :meth:`Flow.index_files`, :meth:`Flow.search_files` instead
    """
    from . import Document

    if read_mode not in {'r', 'rb', None}:
        raise RuntimeError(f'read_mode should be "r", "rb" or None, got {read_mode}')

    def _iter_file_exts(ps):
        return itertools.chain.from_iterable(
            glob.iglob(os.path.expanduser(p), recursive=recursive) for p in ps
        )

    num_docs = 0
    if isinstance(patterns, str):
        patterns = [patterns]
    for g in _iter_file_exts(patterns):
        if os.path.isdir(g):
            continue
        if sampling_rate is None or random.random() < sampling_rate:
            if read_mode is None:
                d = Document(uri=g)
                if to_dataturi:
                    d.convert_uri_to_datauri()
                yield d
            elif read_mode in {'r', 'rb'}:
                with open(g, read_mode) as fp:
                    d = Document(content=fp.read(), uri=g)
                    if to_dataturi:
                        d.convert_uri_to_datauri()
                    yield d
            num_docs += 1
        if size is not None and num_docs >= size:
            break


def from_csv(
    file: Union[str, TextIO],
    field_resolver: Optional[Dict[str, str]] = None,
    size: Optional[int] = None,
    sampling_rate: Optional[float] = None,
    dialect: Union[str, 'csv.Dialect'] = 'excel',
) -> Generator['Document', None, None]:
    """Generator function for CSV. Yields documents.

    :param file: file paths or file handler
    :param field_resolver: a map from field names defined in JSON, dict to the field
            names defined in Document.
    :param size: the maximum number of the documents
    :param sampling_rate: the sampling rate between [0, 1]
    :param dialect: define a set of parameters specific to a particular CSV dialect. could be a string that represents
        predefined dialects in your system, or could be a :class:`csv.Dialect` class that groups specific formatting
        parameters together. If you don't know the dialect and the default one does not work for you,
        you can try set it to ``auto``.
    :yield: documents

    """
    from . import Document

    if hasattr(file, 'read'):
        file_ctx = nullcontext(file)
    else:
        file_ctx = open(file, 'r')

    with file_ctx as fp:
        # when set to auto, then sniff
        try:
            if isinstance(dialect, str) and dialect == 'auto':
                dialect = csv.Sniffer().sniff(fp.read(1024))
                fp.seek(0)
        except:
            dialect = 'excel'  #: can not sniff delimiter, use default dialect

        lines = csv.DictReader(fp, dialect=dialect)
        for value in _subsample(lines, size, sampling_rate):
            if 'groundtruth' in value and 'document' in value:
                yield Document(
                    value['document'], field_resolver=field_resolver
                ), Document(value['groundtruth'], field_resolver=field_resolver)
            else:
                yield Document(value, field_resolver=field_resolver)


def from_huggingface_datasets(
    dataset_path: str,
    field_resolver: Optional[Dict[str, str]] = None,
    size: Optional[int] = None,
    sampling_rate: Optional[float] = None,
    filter_fields: bool = False,
    **datasets_kwargs,
) -> Generator['Document', None, None]:
    """Generator function for Hugging Face Datasets. Yields documents.

    This function helps to load datasets from Hugging Face Datasets Hub
    (https://huggingface.co/datasets) in Jina. Additional parameters can be
    passed to the ``datasets`` library using keyword arguments. The ``load_dataset``
    method from ``datasets`` library is used to load the datasets.

    :param dataset_path: a valid dataset path for Hugging Face Datasets library.
    :param field_resolver: a map from field names defined in ``document`` (JSON, dict) to the field
            names defined in Protobuf. This is only used when the given ``document`` is
            a JSON string or a Python dict.
    :param size: the maximum number of the documents
    :param sampling_rate: the sampling rate between [0, 1]
    :param filter_fields: specifies whether to filter the dataset with the fields
            given in ```field_resolver`` argument.
    :param **datasets_kwargs: additional arguments for ``load_dataset`` method
            from Datasets library. More details at
            https://huggingface.co/docs/datasets/package_reference/loading_methods.html#datasets.load_dataset
    :yield: documents
    """
    from . import Document

    import datasets

    # Load the dataset using given arguments
    data = datasets.load_dataset(dataset_path, **datasets_kwargs)

    # Validate loaded dataset for splits
    if isinstance(data, (datasets.DatasetDict, datasets.IterableDatasetDict)):
        raise ValueError(
            (
                'Please provide a split for dataset using "split" argument. '
                f'The following splits are available for this dataset: {list(data.keys())}'
            )
        )

    # Filter dataset if needed
    if filter_fields:
        if not field_resolver:
            raise ValueError(
                'Filter fields option requires "field_resolver" to be provided.'
            )
        else:
            data.set_format(type=None, columns=list(field_resolver.keys()))

    # Return documents from dataset instances with subsampling if required
    for value in _subsample(data, size, sampling_rate):
        yield Document(value, field_resolver=field_resolver)


def from_ndjson(
    fp: Iterable[str],
    field_resolver: Optional[Dict[str, str]] = None,
    size: Optional[int] = None,
    sampling_rate: Optional[float] = None,
) -> Generator['Document', None, None]:
    """Generator function for line separated JSON. Yields documents.

    :param fp: file paths
    :param field_resolver: a map from field names defined in ``document`` (JSON, dict) to the field
            names defined in Protobuf. This is only used when the given ``document`` is
            a JSON string or a Python dict.
    :param size: the maximum number of the documents
    :param sampling_rate: the sampling rate between [0, 1]
    :yield: documents

    """
    from . import Document

    for line in _subsample(fp, size, sampling_rate):
        value = json.loads(line)
        if 'groundtruth' in value and 'document' in value:
            yield Document(value['document'], field_resolver=field_resolver), Document(
                value['groundtruth'], field_resolver=field_resolver
            )
        else:
            yield Document(value, field_resolver=field_resolver)


def from_lines(
    lines: Optional[Iterable[str]] = None,
    filepath: Optional[str] = None,
    read_mode: str = 'r',
    line_format: str = 'json',
    field_resolver: Optional[Dict[str, str]] = None,
    size: Optional[int] = None,
    sampling_rate: Optional[float] = None,
) -> Generator['Document', None, None]:
    """Generator function for lines, json and csv. Yields documents or strings.

    :param lines: a list of strings, each is considered as a document
    :param filepath: a text file that each line contains a document
    :param read_mode: specifies the mode in which the file
                is opened. 'r' for reading in text mode, 'rb' for reading in binary
    :param line_format: the format of each line ``json`` or ``csv``
    :param field_resolver: a map from field names defined in ``document`` (JSON, dict) to the field
            names defined in Protobuf. This is only used when the given ``document`` is
            a JSON string or a Python dict.
    :param size: the maximum number of the documents
    :param sampling_rate: the sampling rate between [0, 1]
    :yield: documents

    """
    if filepath:
        file_type = os.path.splitext(filepath)[1]
        with open(os.path.expanduser(filepath), read_mode) as f:
            if file_type in _jsonl_ext:
                yield from from_ndjson(f, field_resolver, size, sampling_rate)
            elif file_type in _csv_ext:
                yield from from_csv(f, field_resolver, size, sampling_rate)
            else:
                yield from _subsample(f, size, sampling_rate)
    elif lines:
        if line_format == 'json':
            yield from from_ndjson(lines, field_resolver, size, sampling_rate)
        elif line_format == 'csv':
            yield from from_csv(lines, field_resolver, size, sampling_rate)
        else:
            yield from _subsample(lines, size, sampling_rate)
    else:
        raise ValueError('"filepath" and "lines" can not be both empty')


# https://github.com/ndjson/ndjson.github.io/issues/1#issuecomment-109935996
_jsonl_ext = {'.jsonlines', '.ndjson', '.jsonl', '.jl', '.ldjson'}
_csv_ext = {'.csv', '.tcsv'}


def _sample(iterable, sampling_rate: Optional[float] = None):
    for i in iterable:
        if sampling_rate is None or random.random() < sampling_rate:
            yield i


def _subsample(
    iterable, size: Optional[int] = None, sampling_rate: Optional[float] = None
):
    yield from itertools.islice(_sample(iterable, sampling_rate), size)



from copy import deepcopy

import numpy as np

from docarray import DocumentArray, Document


def test_reduce_doc():
    doc1 = Document(
        text='doc1',
        matches=[Document(id='m0'), Document(id='m2')],
        chunks=[Document(id='c0'), Document(id='c1')],
    )

    doc2 = Document(
        text='doc2',
        embedding=np.zeros(3),
        matches=[Document(id='m0'), Document(id='m1')],
        chunks=[Document(id='c0'), Document(id='c2')],
    )

    DocumentArray._reduce_doc(doc1, doc2)

    assert doc1.text == 'doc1'
    assert (doc1.embedding == np.zeros(3)).all()
    for i in range(3):
        assert f'c{i}' in doc1.chunks
        assert f'm{i}' in doc1.matches


def test_reduce():
    da1, da2 = (
        DocumentArray(
            [
                Document(
                    id='r0',
                    text='da1',
                    matches=[
                        Document(id='r0m0'),
                        Document(id='r0m2'),
                        Document(id='r0m1'),
                    ],
                ),
                Document(
                    id='r2',
                    text='da1',
                    matches=[
                        Document(id='r2m0'),
                        Document(id='r2m2'),
                        Document(id='r2m1'),
                    ],
                ),
            ]
        ),
        DocumentArray(
            [
                Document(
                    id='r0',
                    text='da2',
                    matches=[
                        Document(id='r0m0'),
                        Document(id='r0m1'),
                        Document(id='r0m3'),
                    ],
                ),
                Document(
                    id='r1',
                    text='da2',
                    matches=[
                        Document(id='r1m0'),
                        Document(id='r1m1'),
                        Document(id='r1m2'),
                        Document(id='r1m3'),
                    ],
                ),
                Document(
                    id='r2',
                    text='da2',
                    matches=[
                        Document(id='r2m3'),
                    ],
                ),
            ]
        ),
    )

    da1.reduce(da2)

    for i in range(3):
        assert f'r{i}' in da1
        for j in range(4):
            assert f'r{i}m{j}' in da1[f'r{i}'].matches

    assert da1['r0'].text == 'da1'
    assert da1['r1'].text == 'da2'
    assert da1['r2'].text == 'da1'


def test_reduce_nested():
    da1, da2 = (
        DocumentArray(
            [
                Document(
                    id='r1',
                    chunks=[
                        Document(
                            id='c1',
                            chunks=[Document(id='c1c2')],
                            matches=[Document(id='c1m2')],
                        ),
                    ],
                    matches=[
                        Document(
                            id='m1', chunks=[Document(id='m1c1'), Document(id='m1c2')]
                        ),
                        Document(id='m2'),
                    ],
                ),
            ]
        ),
        DocumentArray(
            [
                Document(
                    id='r1',
                    chunks=[
                        Document(
                            id='c1',
                            chunks=[Document(id='c1c1')],
                            matches=[Document(id='c1m1')],
                        ),
                        Document(
                            id='c2',
                            chunks=[Document(id='c2c1'), Document(id='c2c2')],
                            matches=[Document(id='c2m1'), Document(id='c2m2')],
                        ),
                    ],
                    matches=[
                        Document(
                            id='m1', matches=[Document(id='m1m1'), Document(id='m1m2')]
                        ),
                        Document(
                            id='m2',
                            chunks=[Document(id='m2c1'), Document(id='m2c2')],
                            matches=[Document(id='m2m1'), Document(id='m2m2')],
                        ),
                    ],
                ),
            ]
        ),
    )

    da1.reduce(da2)
    for i in range(1, 3):
        assert f'c{i}' in da1[0].chunks
        assert f'm{i}' in da1[0].matches
        for j in range(1, 3):
            assert f'c{i}c{j}' in da1[0].chunks[f'c{i}'].chunks
            assert f'c{i}m{j}' in da1[0].chunks[f'c{i}'].matches

            assert f'm{i}c{j}' in da1[0].matches[f'm{i}'].chunks
            assert f'm{i}m{j}' in da1[0].matches[f'm{i}'].matches


def test_reduce_mat():
    docs = DocumentArray([Document(id=f'r{i}') for i in range(10)])
    doc_matrix = [deepcopy(docs) for _ in range(10)]
    for i, da in enumerate(doc_matrix):
        for doc in da:
            doc.matches.append(Document(id=str(i)))

    reduced_da = doc_matrix[0].reduce_all(doc_matrix[1:])
    for doc in reduced_da:
        for i in range(10):
            assert str(i) in doc.matches


def test_reduce_data_props():
    da1, da2, da3 = (
        DocumentArray([Document(id='r', text='doc1', chunks=[Document(id='c1')])]),
        DocumentArray(
            [
                Document(
                    id='r',
                    text='doc2',
                    embedding=np.zeros(3),
                    chunks=[Document(id='c2')],
                )
            ]
        ),
        DocumentArray(
            [
                Document(
                    id='r',
                    text='doc3',
                    tags={'a': 'b'},
                    matches=[Document(id='m3')],
                )
            ]
        ),
    )
    da1.reduce_all([da2, da3])
    assert da1[0].text == 'doc1'
    assert da1[0].id == 'r'

    # chunks and matches merged, not overridden
    assert da1[0].chunks[0].id == 'c1'
    assert da1[0].chunks[1].id == 'c2'
    assert da1[0].matches[0].id == 'm3'
    assert da1[0].tags == {'a': 'b'}



import base64
from typing import Optional

from ...helper import random_identity, download_mermaid_url


class PlotMixin:
    """Provide helper functions for :class:`Document` to plot and visualize itself. """

    def _ipython_display_(self):
        """Displays the object in IPython as a side effect"""
        self.summary()

    def summary(self) -> None:
        """ Print non-empty fields and nested structure of this Document object."""
        _str_list = []
        self._plot_recursion(_str_list, indent=0)
        print('\n'.join(_str_list))

    def _plot_recursion(self, _str_list, indent, box_char='├─'):
        prefix = (' ' * indent + box_char) if indent else ''
        _str_list.append(f'{prefix} {self}')

        for a in ('matches', 'chunks'):
            if getattr(self, a):
                prefix = ' ' * (indent + 4) + '└─'
                _str_list.append(f'{prefix} {a}')

                for d in getattr(self, a)[:-1]:
                    d._plot_recursion(_str_list, indent=len(prefix) + 4)
                getattr(self, a)[-1]._plot_recursion(
                    _str_list, indent=len(prefix) + 4, box_char='└─'
                )

    def plot(self):
        """ Plot image data from :attr:`.blob` or :attr:`.uri`. """
        from IPython.display import Image, display

        if self.blob is not None:
            import PIL.Image

            display(PIL.Image.fromarray(self.blob))
        elif self.uri:
            display(Image(self.uri))
        else:
            raise ValueError('`uri` and `blob` is empty')



import pytest

from docarray import DocumentArray

N = 100


def da_and_dam():
    da = DocumentArray.empty(N)
    return (da,)


@pytest.mark.parametrize('da', da_and_dam())
def test_iter_len_bool(da):
    j = 0
    for _ in da:
        j += 1
    assert j == N
    assert j == len(da)
    assert da
    da.clear()
    assert not da


@pytest.mark.parametrize('da', da_and_dam())
def test_repr(da):
    assert f'length={N}' in repr(da)


@pytest.mark.parametrize('da', da_and_dam())
def test_iadd(da):
    oid = id(da)
    dap = DocumentArray.empty(10)
    da += dap
    assert len(da) == N + len(dap)
    nid = id(da)
    assert nid == oid


@pytest.mark.parametrize('da', da_and_dam())
def test_add(da):
    oid = id(da)
    dap = DocumentArray.empty(10)
    da = da + dap
    assert len(da) == N + len(dap)
    nid = id(da)
    assert nid != oid



import numpy as np
import pytest

from docarray import Document
from docarray.score import NamedScore


@pytest.mark.parametrize(
    'doc',
    [
        Document(tags={'hello': 'world', 'sad': {'nest': 123}, 'hello12': 1.2}),
        Document(scores={'hello': NamedScore(value=1.0, description='hello')}),
        Document(location=[1.0, 2.0, 3.0]),
        Document(chunks=[Document()], matches=[Document(), Document()]),
    ],
)
def test_to_from_protobuf(doc):
    docr = Document.from_protobuf(doc.to_protobuf())
    assert docr == doc


def test_to_protobuf():
    with pytest.raises(TypeError):
        Document(text='hello', embedding=np.array([1, 2, 3]), id=1).to_protobuf()

    with pytest.raises(AttributeError):
        Document(tags=1).to_protobuf()

    assert (
        Document(text='hello', embedding=np.array([1, 2, 3])).to_protobuf().text
        == 'hello'
    )
    assert Document(tags={'hello': 'world'}).to_protobuf().tags
    assert len(Document(chunks=[Document(), Document()]).to_protobuf().chunks) == 2



import numpy as np
import pytest
import tensorflow as tf
import torch
from scipy.sparse import csr_matrix, coo_matrix, bsr_matrix, csc_matrix

from docarray import DocumentArray
from docarray.math.ndarray import to_numpy_array
from tests import random_docs


def get_ndarrays_for_ravel():
    a = np.random.random([100, 3])
    a[a > 0.5] = 0
    return [
        (a, False),
        (torch.tensor(a), False),
        (tf.constant(a), False),
        (torch.tensor(a).to_sparse(), True),
        # (tf.sparse.from_dense(a), True),
        (csr_matrix(a), True),
        (bsr_matrix(a), True),
        (coo_matrix(a), True),
        (csc_matrix(a), True),
    ]


@pytest.mark.parametrize('ndarray_val, is_sparse', get_ndarrays_for_ravel())
@pytest.mark.parametrize('target_da', [DocumentArray.empty(100), random_docs(100)])
@pytest.mark.parametrize(
    'protocol', ['protobuf', 'protobuf-array', 'pickle', 'pickle-array']
)
@pytest.mark.parametrize('compress', ['lz4', 'bz2', 'lzma', 'zlib', 'gzip', None])
def test_to_from_bytes(target_da, protocol, compress, ndarray_val, is_sparse):
    bstr = target_da.to_bytes(protocol=protocol, compress=compress)
    print(protocol, compress, len(bstr))
    da2 = DocumentArray.from_bytes(bstr, protocol=protocol, compress=compress)
    assert len(da2) == len(target_da)

    target_da.embeddings = ndarray_val
    target_da.blobs = ndarray_val
    bstr = target_da.to_bytes(protocol=protocol, compress=compress)
    print(protocol, compress, len(bstr))
    da2 = DocumentArray.from_bytes(bstr, protocol=protocol, compress=compress)
    assert len(da2) == len(target_da)

    np.testing.assert_almost_equal(
        to_numpy_array(target_da.embeddings), to_numpy_array(da2.embeddings)
    )
    np.testing.assert_almost_equal(
        to_numpy_array(target_da.blobs), to_numpy_array(da2.blobs)
    )


@pytest.mark.parametrize('target_da', [DocumentArray.empty(100), random_docs(100)])
@pytest.mark.parametrize(
    'protocol', ['protobuf', 'protobuf-array', 'pickle', 'pickle-array']
)
@pytest.mark.parametrize('compress', ['lz4', 'bz2', 'lzma', 'zlib', 'gzip', None])
def test_save_bytes(target_da, protocol, compress, tmpfile):
    target_da.save_binary(tmpfile, protocol=protocol, compress=compress)
    target_da.save_binary(str(tmpfile), protocol=protocol, compress=compress)

    with open(tmpfile, 'wb') as fp:
        target_da.save_binary(fp, protocol=protocol, compress=compress)

    DocumentArray.load_binary(tmpfile, protocol=protocol, compress=compress)
    DocumentArray.load_binary(str(tmpfile), protocol=protocol, compress=compress)
    with open(tmpfile, 'rb') as fp:
        DocumentArray.load_binary(fp, protocol=protocol, compress=compress)


@pytest.mark.parametrize('target_da', [DocumentArray.empty(100), random_docs(100)])
def test_from_to_protobuf(target_da):
    DocumentArray.from_protobuf(target_da.to_protobuf())


@pytest.mark.parametrize('target_da', [DocumentArray.empty(100), random_docs(100)])
def test_from_to_safe_list(target_da):
    DocumentArray.from_list(target_da.to_list())



import numpy as np
import paddle
import pytest
import tensorflow as tf
import torch
from scipy.sparse import csr_matrix, coo_matrix, bsr_matrix, csc_matrix

from docarray import Document


def get_ndarrays():
    a = np.random.random([10, 3])
    a[a > 0.5] = 0
    return [
        (a, False),
        (torch.tensor(a), False),
        (tf.constant(a), False),
        (paddle.to_tensor(a), False),
        (torch.tensor(a).to_sparse(), True),
        (tf.sparse.from_dense(a), True),
        (csr_matrix(a), True),
        (bsr_matrix(a), True),
        (coo_matrix(a), True),
        (csc_matrix(a), True),
    ]


@pytest.mark.parametrize('ndarray_val, is_sparse', get_ndarrays())
@pytest.mark.parametrize('attr', ['embedding', 'blob'])
def test_ndarray_force_numpy(ndarray_val, attr, is_sparse):
    d = Document()
    setattr(d, attr, ndarray_val)
    assert type(getattr(Document.from_protobuf(d.to_protobuf()), attr)) is type(
        ndarray_val
    )
